import xarray as xa 
import pandas as pd
import numpy as np
from tqdm import tqdm
try:
    import matplotlib.pyplot as plt
except ImportError:
    print('skipping matplotlib')
import sys
from functools import partial
import time
try:
    from dask.diagnostics import ProgressBar
except ImportError:
    print('skipping dask import')

from xmip.preprocessing import rename_cmip6, promote_empty_dims, broadcast_lonlat, correct_coordinates
import StraitFlux.preprocessing as prepro
import StraitFlux.functions as func
from StraitFlux.indices import check_availability_indices, prepare_indices




def transports(product,strait,model,time_start,time_end,file_u,file_v,file_t,file_z, file_zu=None, file_zv=None,mesh_dxv=0, mesh_dyu=0,coords=0,set_latlon=False,lon_p=0,lat_p=0,file_s='',file_sic='',file_sit='',Arakawa='',rho=1026,cp=3996, Tref=0,path_save='',path_indices='',path_mesh='',saving=True):

    '''Calculation of Transports using line integration

    INPUT Parameters:
    product (str): volume, heat, salt or ice
    strait (str): desired oceanic strait, either pre-defined from indices file or new
    model (str): desired CMIP6 model or reanalysis
    time_start (str or int): starting year
    time_end (str or int): ending year
    file_u (str OR ): path + filename(s) of u field(s); use ice velocities (ui) for ice transports; (multiple files possible, use *; must be possible to combine files over time coordinate)
    file_v (str): path + filename(s) of v field(s); use ice velocities (vi) for ice transports; (multiple files possible, use *)
    file_t (str): path + filename(s) of temperature field(s); (multiple files possible, use *)
    file_z (str): path + filename(s) of cell thickness field(s); (multiple files possible, use *)

    OPTIONAL:
    mesh_dxu/mesh_dyv (array): arrays containing the exact grid cell dimensions at northern and eastern grid cell faces of u and v cells (dxv and dyu); if not supplied will be calculated
    coords (tuple): coordinates for strait, if not pre-defined: (latitude_start,longitude_start,latitude_end,longitude_end)
    set_latlon: set True if you wish to pass arrays of latitudes and longitudes
    lon (array): longitude coordinates for strait, if not pre-defined. (range -180 to 180; same length as lat needed!)
    lat (array): latitude coordinates for strait, if not pre-defined. (range -90 to 90; same length as lon needed!)
    file_s (str): only needed for salinity transports; path + filename(s) of salinity field(s); (multiple files possible, use *)
    file_sic (str): only needed for ice transports; path + filename(s) of sea ice concentration field(s); (multiple files possible, use *)
    file_sit (str): only needed for ice transports; path + filename(s) of sea ice thickness field(s); (multiple files possible, use *)
    Arakawa (str): Arakawa-A, Arakawa-B or Arakawa-C; only needed if automatic check fails
    rho (int or array): default = 1026 kg/m3
    cp (int or array): default = 3996 J/(kgK)
    Tref (int or array): default = 0°C
    path_save (str): path to save transport data
    path_indices (str): path to save indices data
    path_mesh (str): path to save mesh data


    RETURNS:
    volume, heat, salt or ice transports through specified strait for specified model

    '''


    if product == 'ice':
        partial_func = partial(prepro._preprocess1i)
    else:
        partial_func = partial(prepro._preprocess1)


    try:
        indices=xa.open_dataset(path_indices+model+'_'+strait+'_indices.nc')
    except OSError:
        print('calc indices')
        print('read and load files for indices')
        if product == 'ice':
            ti = xa.open_mfdataset(file_sit, preprocess=partial_func).isel(time=0)
        else:
            ti = xa.open_mfdataset(file_t, preprocess=partial_func).isel(time=0)
        ui = xa.open_mfdataset(file_u, preprocess=partial_func).isel(time=0)
        vi = xa.open_mfdataset(file_v, preprocess=partial_func).isel(time=0)
        try:
            with ProgressBar():
                ti=ti.load()
                ui=ui.load()
                vi=vi.load()
        except NameError:
            ti=ti.load()
            ui=ui.load()
            vi=vi.load()
        indices,line = check_availability_indices(ti,strait,model,coords,lon_p,lat_p,set_latlon)
        i2=indices.indices.where(indices.indices!=0)
        try:
            if product == 'ice':
                plt.pcolormesh(ti.sithick/ti.sithick,cmap='tab20c')
            else:
                plt.pcolormesh((ti.thetao/ti.thetao),cmap='tab20c')
            plt.scatter(i2[:,2],i2[:,3],color='tab:red',s=0.1,marker='x')
            plt.scatter(i2[:,0],i2[:,1],color='tab:red',s=0.1,marker='x')
            plt.title(model+'_'+strait,fontsize=14)
            plt.ylabel('y',fontsize=14)
            plt.xlabel('x',fontsize=14)
            plt.savefig(path_save+strait+'_'+model+'_indices.png')
            plt.close()
        except:
            print('skipping Plot')
        out_u,out_v,out_u_vz = prepare_indices(indices)
        if product == 'ice':
            func.check_indices(indices,out_u,out_v,ti,ui,vi,strait,model,path_save)
        else:
            func.check_indices(indices,out_u,out_v,ti,ui,vi,strait,model,path_save)
        if saving == True:
            indices.to_netcdf(path_indices+model+'_'+strait+'_indices.nc')

    out_u,out_v,out_u_vz = prepare_indices(indices)
    if Arakawa in ['Arakawa-A','Arakawa-B','Arakawa-C']:
        grid=Arakawa
    elif Arakawa == '':
        try:
            file = open(path_mesh+model+'grid.txt', 'r')
            grid= file.read()
        except OSError:
            try:
                grid = func.check_Arakawa(ui,vi,ti,model)
                if saving == True:
                    with open(path_mesh+model+'grid.txt', 'w') as f:
                        f.write(grid)
            except NameError:
                print('read and load files for grid check')
                ti = xa.open_mfdataset(file_t, preprocess=partial_func).isel(time=0)
                ui = xa.open_mfdataset(file_u, preprocess=partial_func).isel(time=0)
                vi = xa.open_mfdataset(file_v, preprocess=partial_func).isel(time=0)
                try:
                    with ProgressBar():
                        ti=ti.load()
                        ui=ui.load()
                        vi=vi.load()
                except NameError:
                    ti=ti.load()
                    ui=ui.load()
                    vi=vi.load()
                grid = func.check_Arakawa(ui,vi,ti,model)
                if saving == True:
                    with open(path_mesh+model+'grid.txt', 'w') as f:
                        f.write(grid)
    else:
        print('grid not known')
        sys.exit()


    min_x=np.nanmin((min(out_u[:,0],default=np.nan),min(out_v[:,0],default=np.nan)))
    max_x=np.nanmax((max(out_u[:,0],default=np.nan),max(out_v[:,0],default=np.nan)))
    min_y=np.nanmin((min(out_u[:,1],default=np.nan),min(out_v[:,1],default=np.nan)))
    max_y=np.nanmax((max(out_u[:,1],default=np.nan),max(out_v[:,1],default=np.nan)))

    if min_x == -1:
        min_x = 0
        max_x = max_x + 1
    partial_func2 = partial(prepro._preprocess2,lon_bnds=(int(min_x)-1,int(max_x)+1),lat_bnds=(int(min_y)-1,int(max_y)+1))
    try:
        mu=xa.open_mfdataset(path_mesh+'mesh_dyu_'+model+'.nc', preprocess=partial_func2)
        mv=xa.open_mfdataset(path_mesh+'mesh_dxv_'+model+'.nc', preprocess=partial_func2)
    except:
        if isinstance(mesh_dxv, xa.Dataset):
            original_var_name = list(mesh_dxv.data_vars)[0]
            mesh_dxv.rename({original_var_name: "dxv"}).to_netcdf(path_mesh+'mesh_dxv_'+model+'.nc')
            mesh_dyu.rename({original_var_name: "dyu"}).to_netcdf(path_mesh+'mesh_dyu_'+model+'.nc')
            mu=xa.open_mfdataset(path_mesh+'mesh_dyu_'+model+'.nc', preprocess=partial_func2)
            mv=xa.open_mfdataset(path_mesh+'mesh_dxv_'+model+'.nc', preprocess=partial_func2)
        elif isinstance(mesh_dxv, xa.DataArray):
            mesh_dxv.to_dataset(name="dxv").to_netcdf(path_mesh+'mesh_dxv_'+model+'.nc')
            mesh_dyu.to_dataset(name='dyu').to_netcdf(path_mesh+'mesh_dyu_'+model+'.nc')
            mu=xa.open_mfdataset(path_mesh+'mesh_dyu_'+model+'.nc', preprocess=partial_func2)
            mv=xa.open_mfdataset(path_mesh+'mesh_dxv_'+model+'.nc', preprocess=partial_func2)
        else:       
            print('meshes not supplied or not DataArray/DataSet: calc horizontal meshes')
            try:
                mu,mv = prepro.calc_dxdy(model,ui,vi,path_mesh)
                mu=xa.open_mfdataset(path_mesh+'mesh_dyu_'+model+'.nc', preprocess=partial_func2)
                mv=xa.open_mfdataset(path_mesh+'mesh_dxv_'+model+'.nc', preprocess=partial_func2) 
            except NameError:
                print('read and load files for mesh')
                ui = xa.open_mfdataset(file_u, preprocess=partial_func).isel(time=0)
                vi = xa.open_mfdataset(file_v, preprocess=partial_func).isel(time=0)
                try:
                    with ProgressBar():
                        ui=ui.load()
                        vi=vi.load()
                except NameError:
                    ui=ui.load()
                    vi=vi.load()
                mu,mv = prepro.calc_dxdy(model,ui,vi,path_mesh)
                mu=xa.open_mfdataset(path_mesh+'mesh_dyu_'+model+'.nc', preprocess=partial_func2)
                mv=xa.open_mfdataset(path_mesh+'mesh_dxv_'+model+'.nc', preprocess=partial_func2) 


    print('read t, u and v fields')
    partial_func = partial(prepro._preprocess2,lon_bnds=(int(min_x)-1,int(max_x)+1),lat_bnds=(int(min_y)-1,int(max_y)+1))
    t = xa.open_mfdataset(file_t, preprocess=partial_func,chunks={'time':1})
    u = xa.open_mfdataset(file_u, preprocess=partial_func,chunks={'time':1})
    v = xa.open_mfdataset(file_v, preprocess=partial_func,chunks={'time':1})
    if product == 'ice':
        sit = xa.open_mfdataset(file_sit, preprocess=partial_func,chunks={'time':1})
        sic = xa.open_mfdataset(file_sic, preprocess=partial_func,chunks={'time':1})
    if 'time' in t.dims and t.dims['time'] > 1:
        t=t.sel(time=slice(str(time_start),str(time_end)))
        u=u.sel(time=slice(str(time_start),str(time_end)))
        v=v.sel(time=slice(str(time_start),str(time_end)))
        if product == 'ice':
            sit=sit.sel(time=slice(str(time_start),str(time_end)))
            sic=sic.sel(time=slice(str(time_start),str(time_end)))
    elif 'time' not in t.dims:
        t=t.expand_dims(dim={"time": 1})
        u=u.expand_dims(dim={"time": 1})
        v=v.expand_dims(dim={"time": 1})
        if product == 'ice':
            sit=sit.expand_dims(dim={"time": 1})
            sic=sic.expand_dims(dim={"time": 1})

    if product in ['volume','heat','salt']:
        deltaz = xa.open_mfdataset(file_z, preprocess=partial_func,chunks={'time':1})[['thkcello']]
        if 'time' in deltaz.dims:
            deltaz=deltaz.sel(time=slice(str(time_start),str(time_end)))
    mu=mu.sel(x=slice(int(min_x)-1,int(max_x)+1),y=slice(int(min_y)-1,int(max_y)+1)).load()
    mv=mv.sel(x=slice(int(min_x)-1,int(max_x)+1),y=slice(int(min_y)-1,int(max_y)+1)).load()

    print('load t, u and v fields')
    try:
        with ProgressBar():
            t=t.load()
            u=u.load()
            v=v.load()
            if product in ['volume','heat','salt']:
                deltaz=deltaz.load()
    except NameError:
        t=t.load()
        u=u.load()
        v=v.load()
        if product in ['volume','heat','salt']:
            deltaz=deltaz.load()

    if product in ['volume','heat','salt']:
        # Check if file_zu and file_zv are provided
        if file_zu and file_zv:
            try:
                dzu3 = xa.open_mfdataset(file_zu, preprocess=partial_func, chunks={'time': 1})[['thkcello']]
                dzv3 = xa.open_mfdataset(file_zv, preprocess=partial_func, chunks={'time': 1})[['thkcello']]
            except Exception as e:
                print(f"Error opening file_zu or file_zv: {e}")
                dzu3, dzv3 = func.calc_dz_faces(deltaz, grid, model, path_mesh)
        else:
            dzu3, dzv3 = func.calc_dz_faces(deltaz, grid, model, path_mesh)
        
    trans = xa.Dataset({'tot_'+product+'_flux':(('time'),np.array(np.zeros(t.time.size)))},coords=dict(time=t.time))
    sign_v=[]
    indi=indices.indices[:,2][indices.indices[:,3]!=0]
    for ind in range(len(indi)-1):
        if indi[ind]<indi[ind+1]:
            sign_v=np.append(sign_v,1)
        elif indi[ind]>=indi[ind+1]:
            if indi[ind+1] in [1,0,-1]:
                sign_v=np.append(sign_v,1)
            else:
                sign_v=np.append(sign_v,-1)
    try:
        sign_v=np.append(sign_v,sign_v[-1])
    except IndexError:
        pass


    print(' ...calculating transport')
    trans_arr = []
    udata = u.uo
    vdata = v.vo
    Tdata = t

    if product == 'salt':
        Sdata = xa.open_mfdataset(file_s, preprocess=partial_func,chunks={'time':1}).sel(time=slice(str(time_start),str(time_end)))


    if product in ['volume','heat','salt']:
        udata,vdata2,dzu3,dzv3,mu2,mv2 = func.transform_Arakawa(grid,mu,mv,deltaz,dzu3,dzv3,udata,vdata)


    if product == 'volume':
        print('calc u')
        udata=udata*mu2.dyu.values*dzu3.values
        print('calc v')
        vdata=vdata*mv2.dxv.values*dzv3.values


    if product == 'heat':
        print('rolling T')
        Tudata = func.interp_TS(Tdata.thetao,'x')
        Tvdata = func.interp_TS(Tdata.thetao,'y')
        print('calc u')
        udata=udata*mu2.dyu.values*dzu3.values*(Tudata.values-Tref)
        print('calc v')
        vdata=vdata*mv2.dxv.values*dzv3.values*(Tvdata.values-Tref)



    if product == 'salt':
        print('rolling S')
        Sudata = func.interp_TS(Sdata.so,'x')
        Svdata = func.interp_TS(Sdata.so,'y')
        print('calc u')
        udata=udata*mu2.dyu.values*dzu3.values*Sudata.values
        print('calc v')
        vdata=vdata*mv2.dxv.values*dzv3.values*Svdata.values

    if product == 'ice':
        if sic.siconc.max() > 1:
            sic['siconc'] = sic.siconc/100
        print('calc u')
        udata=udata*mu.dyu.values*sit.sithick.values*sic.siconc.values
        print('calc v')
        vdata=vdata*mv.dxv.values*sit.sithick.values*sic.siconc.values

    udata = udata.fillna(0.)
    vdata = vdata.fillna(0.)
    print('calc line')
    if product in ['volume','heat','salt']:
        udata = udata.sum(dim='lev')
        vdata = vdata.sum(dim='lev')
    datau = xa.Dataset({'inte':(('time','y','x'),udata.data)},coords=({'time':('time',udata.time.data),'x':('x',udata.x.data),'y':('y',udata.y.data)}))
    datav = xa.Dataset({'inte':(('time','y','x'),vdata.data)},coords=({'time':('time',vdata.time.data),'x':('x',vdata.x.data),'y':('y',vdata.y.data)}))
    pointsu = np.zeros(datau.inte.shape)
    data1u = datau.inte

    for l in range(len(out_u)):
        if out_u_vz[l][2] == -1:
            pointsu[:,int(out_u_vz[l,1]-min_y+1),int(out_u_vz[l,0]-min_x+1)] = data1u[:,int(out_u_vz[l,1]-min_y+1),int(out_u_vz[l,0]-min_x+1)] * (-1)
        else:
            pointsu[:,int(out_u_vz[l,1]-min_y+1),int(out_u_vz[l,0]-min_x+1)] = data1u[:,int(out_u_vz[l,1]-min_y+1),int(out_u_vz[l,0]-min_x+1)]

    pointsv = np.zeros(datav.inte.shape)
    data1v = datav.inte

    indi1=indices.indices[:,2][indices.indices[:,3]!=0]
    indi2=indices.indices[:,3][indices.indices[:,3]!=0]

    for m in range(len(indi1)-1):
        pointsv[:,int(indi2[m]-min_y+1),int(indi1[m]-min_x+1)] = data1v[:,int(indi2[m]-min_y+1),int(indi1[m]-min_x+1)] * sign_v[m]

    vp = xa.Dataset({'v':(('time','y','x'),pointsv)},coords=datav.coords)
    up = xa.Dataset({'u':(('time','y','x'),pointsu)},coords=datau.coords)
    ges_l = datau.copy()
    ges_l['inte'] = vp['v'] + up['u']
    if product == 'heat':
        ges_l['inte'] = ges_l['inte'] * rho * cp
    elif product == 'salt':
        ges_l['inte'] = ges_l['inte'] * rho
    #ges_l.to_netcdf(model+'_'+strait+'_test.nc')
    summ = ges_l.sum(dim=['x','y'])
    summ = summ.inte
    trans_arr = np.append(trans_arr,summ)

    trans[model]= (['time'],trans_arr)
    trans = trans.drop_vars('tot_'+product+'_flux')
    trans.to_netcdf(path_save+strait+'_'+product+'_'+model+'_'+str(time_start)+'-'+str(time_end)+'.nc')

    return trans






import xarray as xa
import numpy as np
try:
    import matplotlib.pyplot as plt
except ImportError:
    print('skipping matplotlib')
from tqdm import tqdm
import sys
try:
    from dask.diagnostics import ProgressBar
except ImportError:
    print('skipping dask import')
from StraitFlux.indices import check_availability_indices, prepare_indices

def check_Arakawa(u_data,v_data,T_data,model):

    u=u_data
    v=v_data
    t=T_data

    print('checking grid')
    if model in ['MPI-ESM1-2-LR','MPI-ESM1-2-HR']:
        grid='Arakawa-C'
        print(grid)
    elif u.lat[int(len(t.y)/2),0].values == t.lat[int(len(t.y)/2),0].values and v.lat[int(len(t.y)/2),0].values != t.lat[int(len(t.y)/2),0].values:
        if v.lon[int(len(t.y)/2),0].values == t.lon[int(len(t.y)/2),0].values and u.lon[int(len(t.y)/2),0].values != t.lon[int(len(t.y)/2),0].values:
            grid='Arakawa-C'
            print(grid)
        else:
            print('grid not recognized, check manually')
            sys.exit()
    elif u.lat[int(len(t.y)/2),0].values == v.lat[int(len(t.y)/2),0].values and u.lon[int(len(t.y)/2),0].values == v.lon[int(len(t.y)/2),0].values:
        if u.lat[int(len(t.y)/2),0].values != t.lat[int(len(t.y)/2),0].values and u.lon[int(len(t.y)/2),0].values != t.lon[int(len(t.y)/2),0].values:
            grid='Arakawa-B'
            print(grid)
        elif u.lat[int(len(t.y)/2),0].values == t.lat[int(len(t.y)/2),0].values and u.lon[int(len(t.y)/2),0].values == t.lon[int(len(t.y)/2),0].values:
            grid='Arakawa-A'
            print(grid)
        elif u.lat[int(len(t.y)/2),0].values == t.lat[int(len(t.y)/2),0].values == v.lat[int(len(t.y)/2),0].values and u.lon[int(len(t.y)/2),0].values != t.lon[int(len(ti.y)/2),0].values != v.lon[int(len(ti.y)/2),0].values:
            grid='Arakawa-E'
            print(grid+'?')
        else:
            print('grid not recognized, check manually')
            sys.exit()
    else:
        print('grid not recognized, check manually')
        sys.exit()
    return grid

def transform_Arakawa(grid,mu,mv,deltaz,dzu3,dzv3,udata,vdata):

    deltaz2=deltaz.thkcello###.mean(dim='time')
    dzs=deltaz2.sum(dim='lev').where(deltaz2.sum(dim='lev')!=0) #axis=0

    if grid == 'Arakawa-C':
        mu2=mu
        mv2=mv
    if grid == 'Arakawa-B':
        print('transforming Arakawa-B to Arakawa-C')
        dzuv=deltaz2.rolling(x=2,min_periods=1).mean().rolling(y=2,min_periods=1).mean()
        dzuv2=dzuv.cumsum('lev').where(dzuv.cumsum('lev')<=dzs) #dzuvs
        dzuv2=dzuv2.fillna(dzs)#dzuvs
        dzuv3=dzuv2.copy(data=np.diff(dzuv2, axis=dzuv2.get_axis_num("lev"), prepend=0))
        mu2=mu.rolling(x=2,min_periods=1).mean().rolling(y=2,min_periods=1).mean()
        mv2=mv.rolling(y=2,min_periods=1).mean().rolling(x=2,min_periods=1).mean()
        udata2=(udata*mu2.dyu.values*dzuv3.values).rolling(y=2,min_periods=1).mean()/(mu.dyu.values*dzu3.values)#.fillna(0)
        vdata2=(vdata*mv2.dxv.values*dzuv3.values).rolling(x=2,min_periods=1).mean()/(mv.dxv.values*dzv3.values)
        udata2=udata2.where(udata2>-2).where(udata2<2).fillna(0)
        vdata2=vdata2.where(vdata2>-2).where(vdata2<2).fillna(0)
        udata=udata2*(udata/udata).values
        vdata=vdata2*(udata/udata).values


    if grid == 'Arakawa-A':
        print('transforming Arakawa-A to Arakawa-C')
        mu2=mu.rolling(x=2,min_periods=1).mean()#.rolling(y=2,min_periods=1).mean()
        mv2=mv.rolling(y=2,min_periods=1).mean()#.rolling(x=2,min_periods=1).mean()
        print('equation to get u/v at T faces')
        udata2=(udata*mu.dyu.values*deltaz2.values).rolling(x=2,min_periods=1).mean()/(mu2.dyu.values*dzu3.values)#.fillna(0)
        vdata2=(vdata*mv.dxv.values*deltaz2.values).rolling(y=2,min_periods=1).mean()/(mv2.dxv.values*dzv3.values)
        udata=udata2.where(udata2>-1000).where(udata2<1000).fillna(0)
        vdata=vdata2.where(vdata2>-1000).where(vdata2<1000).fillna(0)

    return udata,vdata,dzu3,dzv3,mu2,mv2

def check_indices(indices,out_u,out_v,t,u,v,strait,model,path_save):
    lp=indices.indices[-1][indices.indices[-1] != 0].values
    slp=indices.indices[-2][indices.indices[-2] != 0].values
    fp=indices.indices[0][indices.indices[0] != 0].values
    sfp=indices.indices[1][indices.indices[1] != 0].values
    tfp=indices.indices[2][indices.indices[2] != 0].values
    #last point:
    if indices.indices[-1][0] == 0 and indices.indices[-1][1] == 0:
        if v.vo[int(lp[1]-1),int(lp[0]-1)].values > 0 or v.vo[int(lp[1]-1),int(lp[0]-1)].values < 0:
            if v.vo[int(slp[1]-1),int(slp[0]-1)].values > 0 or v.vo[int(slp[1]-1),int(slp[0]-1)].values < 0:
                print('!!!ATTENTION!!!: last point water, recheck indices line!')
            else:
                print('dropping last point...')
        else:
            print('line good')
    else:
        if u.uo[int(lp[1]-1),int(lp[0]-1)].values > 0 or u.uo[int(lp[1]-1),int(lp[0]-1)].values < 0:
            if u.uo[int(slp[1]-1),int(slp[0]-1)].values > 0 or u.uo[int(slp[1]-1),int(slp[0]-1)].values < 0:
                print('!!!ATTENTION!!!: last point water, recheck indices line!')
            else:
                print('dropping last point...')
        else:
            print('line good')
            

    #first point:
    if indices.indices[0][2] == 0 and indices.indices[0][3] == 0:
        if u.uo[int(fp[1]-1),int(fp[0]-1)].values > 0 or u.uo[int(fp[1]-1),int(fp[0]-1)].values < 0:
            if u.uo[int(sfp[1]-1),int(sfp[0]-1)].values > 0 or u.uo[int(sfp[1]-1),int(sfp[0]-1)].values < 0:
                print('!!!ATTENTION!!!: first point water, recheck indices line!')
            else:
                print('dropping last point...')
        else:
            print('line good')
    else:
        if v.vo[int(fp[1]-1),int(fp[0]-1)].values > 0 or v.vo[int(fp[1]-1),int(fp[0]-1)].values < 0:
            if v.vo[int(sfp[1]-1),int(sfp[0]-1)].values > 0 or v.vo[int(sfp[1]-1),int(sfp[0]-1)].values < 0:
                print('!!!ATTENTION!!!: first point water, recheck indices line!')
            else:
                print('dropping last point...')
        else:
            print('line good')
       
    
    out_u,out_v,out_u_vz = prepare_indices(indices)

    min_x=np.nanmin((min(out_u[:,0],default=np.nan),min(out_v[:,0],default=np.nan)))
    max_x=np.nanmax((max(out_u[:,0],default=np.nan),max(out_v[:,0],default=np.nan)))
    min_y=np.nanmin((min(out_u[:,1],default=np.nan),min(out_v[:,1],default=np.nan)))
    max_y=np.nanmax((max(out_u[:,1],default=np.nan),max(out_v[:,1],default=np.nan)))

    if min_x == -1:
        min_x = 0
        max_x = max_x + 1
        
    t=t.sel(x=slice(int(min_x)-2,int(max_x)+2),y=slice(int(min_y)-2,int(max_y)+2)).load()
    u=u.sel(x=slice(int(min_x)-1,int(max_x)+1),y=slice(int(min_y)-1,int(max_y)+1)).load()
    v=v.sel(x=slice(int(min_x)-1,int(max_x)+1),y=slice(int(min_y)-1,int(max_y)+1)).load()
    try:
        plt.title(model+'_'+strait,fontsize=14)
        try:
            plt.pcolormesh(t.x,t.y,(t.thetao/t.thetao),cmap='tab20c')
        except AttributeError:
            plt.pcolormesh(t.x,t.y,(t.sithick/t.sithick),cmap='tab20c')
        plt.scatter(out_v[:,0],out_v[:,1]+0.5,marker='_',c='r',s=200)
        plt.scatter(out_u[:,0]+0.5,out_u[:,1],marker='|',c='r',s=200)
        plt.ylabel('y',fontsize=14)
        plt.xlabel('x',fontsize=14)
        plt.savefig(path_save+strait+'_'+model+'_indices_check.png')
        plt.close()
    except:
        print('skipping Plot')

def interp_TS(ds,d):
    return ds.rolling({d:2},min_periods=1).mean()

def calc_dz_faces(deltaz,grid,model,path_mesh):

    if model in ['MPI-ESM1-2-LR','MPI-ESM1-2-HR']:
        print('swap')
        deltaz['y']=np.arange(len(deltaz.y)-1,-1,-1)
        deltaz=deltaz.sortby('y')
    try:
        with ProgressBar():
            zv=deltaz.thkcello.values
    except NameError:
        zv=deltaz.thkcello.values
    z0u=np.zeros(np.shape(zv))
    z0v=np.zeros(np.shape(zv))
    print('calc dz at cell faces')
    if 'time' in deltaz.dims:
        deltazu=xa.Dataset({'thkcello':(('time','lev','y','x'),np.zeros(np.shape(zv)))},coords=({'time':('time',deltaz.time.data),'lev':('lev',deltaz.lev.data),'x':('x',deltaz.x.data),'y':('y',deltaz.y.data)}))
        deltazv=xa.Dataset({'thkcello':(('time','lev','y','x'),np.zeros(np.shape(zv)))},coords=({'time':('time',deltaz.time.data),'lev':('lev',deltaz.lev.data),'x':('x',deltaz.x.data),'y':('y',deltaz.y.data)}))
    else:
        deltazu=xa.Dataset({'thkcello':(('lev','y','x'),np.zeros(np.shape(zv)))},coords=({'lev':('lev',deltaz.lev.data),'x':('x',deltaz.x.data),'y':('y',deltaz.y.data)}))
        deltazv=xa.Dataset({'thkcello':(('lev','y','x'),np.zeros(np.shape(zv)))},coords=({'lev':('lev',deltaz.lev.data),'x':('x',deltaz.x.data),'y':('y',deltaz.y.data)}))


    if grid == 'Arakawa-C':
        if 'time' in deltaz.dims:
            for i in tqdm(range(len(deltaz.y)-1)):
                #print(i)
                for j in range(len(deltaz.x)-1):
                    p=np.nansum(zv[:,:,i,j:j+2],axis=1).argmin(axis=1)
                    for k in range(len(p)):
                        if p[k] == 0:
                            z0u[k,:,i,j]=zv[k,:,i,j]
                        if p[k] ==1:
                            l=np.isnan(zv[k,:,i,j+1]).argmax(axis=0)-1
                            z0u[k,:l,i,j]=zv[k,:l,i,j]
                            z0u[k,l,i,j]=zv[k,l,i,j+1]
                            if l >= 0:
                                z0u[k,l+1:,i,j]=np.nan
            deltazu['thkcello'][:,:,:,:]=z0u
            deltazu['thkcello'][:,:,:,-1]=zv[:,:,:,-1]
            deltazu['thkcello'][:,:,-1,:]=zv[:,:,-1,:]

            for j in tqdm(range(len(deltaz.x)-1)):
                #print(j)
                for i in range(len(deltaz.y)-1):
                    p=np.nansum(zv[:,:,i:i+2,j],axis=1).argmin(axis=1)
                    for k in range(len(p)):
                        if p[k] == 0:
                            z0v[k,:,i,j]=zv[k,:,i,j]
                        if p[k] ==1:
                            l=np.isnan(zv[k,:,i+1,j]).argmax(axis=0)-1
                            z0v[k,:l,i,j]=zv[k,:l,i,j]
                            z0v[k,l,i,j]=zv[k,l,i+1,j]
                            if l >= 0:
                                z0v[k,l+1:,i,j]=np.nan

            deltazv['thkcello'][:,:,:,:]=z0v
            deltazv['thkcello'][:,:,:,-1]=zv[:,:,:,-1]
            deltazv['thkcello'][:,:,-1,:]=zv[:,:,-1,:]

        else:
            for i in tqdm(range(len(deltaz.y)-1)):
                #print(i)
                for j in range(len(deltaz.x)-1):
                    p=np.nansum(zv[:,i,j:j+2],axis=0).argmin()
                    if p == 0:
                        z0u[:,i,j]=zv[:,i,j]
                    if p ==1:
                        z0u[:,i,j]=zv[:,i,j+1]
            deltazu['thkcello'][:,:,:]=z0u
            deltazu['thkcello'][:,:,-1]=zv[:,:,-1]
            deltazu['thkcello'][:,-1,:]=zv[:,-1,:]

            for j in tqdm(range(len(deltaz.x)-1)):
                for i in range(len(deltaz.y)-1):
                    p=np.nansum(zv[:,i:i+2,j],axis=0).argmin()
                    if p == 0:
                        z0v[:,i,j]=zv[:,i,j]
                    if p ==1:
                        z0v[:,i,j]=zv[:,i+1,j]
            deltazv['thkcello'][:,:,:]=z0v
            deltazv['thkcello'][:,:,-1]=zv[:,:,-1]
            deltazv['thkcello'][:,-1,:]=zv[:,-1,:]

    elif grid in ['Arakawa-B','Arakawa-A']:
        if 'time' in deltaz.dims:
            for k in tqdm(range(len(deltaz.time))):
                #print(k)
                zv=deltaz.thkcello[k].values
                z0u=np.zeros(np.shape(zv))
                z0v=np.zeros(np.shape(zv))
                for i in range(len(deltaz.y)-1):
                    #print(i)
                    for j in range(len(deltaz.x)-1):
                        p=np.argwhere(np.nansum(zv[:,i:i+2,j:j+2],axis=0) == np.min(np.nansum(zv[:,i:i+2,j:j+2],axis=0)))[0]
                        if p[0] == 0:
                            if p[1] == 0:
                                z0u[:,i,j]=zv[:,i,j]
                            elif p[1] == 1:
                                z0u[:,i,j]=zv[:,i,j+1]
                        elif p[0] == 1:
                            if  p[1] == 0:
                                z0u[:,i,j]=zv[:,i+1,j]
                            elif p[1] == 1:
                                z0u[:,i,j]=zv[:,i+1,j+1]
                deltazu['thkcello'][k,:,:,:]=z0u
                deltazu['thkcello'][k,:,:,-1]=zv[:,:,-1]
                deltazu['thkcello'][k,:,-1,:]=zv[:,-1,:]
                deltazv['thkcello'][k,:,:,:]=z0u
                deltazv['thkcello'][k,:,:,-1]=zv[:,:,-1]
                deltazv['thkcello'][k,:,-1,:]=zv[:,-1,:]


        else:
            for i in tqdm(range(len(deltaz.y)-1)):
                #print(i)
                for j in range(len(deltaz.x)-1):
                    p=np.argwhere(np.nansum(zv[:,i:i+2,j:j+2],axis=0) == np.min(np.nansum(zv[:,i:i+2,j:j+2],axis=0)))[0]
                    if p[0] == 0:
                        if p[1] == 0:
                            z0u[:,i,j]=zv[:,i,j]
                        elif p[1] == 1:
                            z0u[:,i,j]=zv[:,i,j+1]
                    elif p[0] == 1:
                        if  p[1] == 0:
                            z0u[:,i,j]=zv[:,i+1,j]
                        elif p[1] == 1:
                            z0u[:,i,j]=zv[:,i+1,j+1]
            deltazu['thkcello'][:,:,:]=z0u
            deltazu['thkcello'][:,:,-1]=zv[:,:,-1]
            deltazu['thkcello'][:,-1,:]=zv[:,-1,:]
            deltazv['thkcello'][:,:,:]=z0u
            deltazv['thkcello'][:,:,-1]=zv[:,:,-1]
            deltazv['thkcello'][:,-1,:]=zv[:,-1,:]

    if model in ['MPI-ESM1-2-LR','MPI-ESM1-2-HR']:
        print('swap')
        deltazu['y']=np.arange(len(deltazu.y)-1,-1,-1)
        deltazu=deltazu.sortby('y')
        deltazv['y']=np.arange(len(deltazv.y)-1,-1,-1)
        deltazv=deltazv.sortby('y')

    return deltazu.thkcello,deltazv.thkcello



import xarray as xa
import numpy as np
from tqdm import tqdm
from xmip.preprocessing import rename_cmip6,promote_empty_dims, broadcast_lonlat, correct_coordinates

def renaming_dict_exp():
    rename_dict = {
        # dim labels (order represents the priority when checking for the dim labels)
        "x": ["i", "ni", "xh", "nlon","ncl3"],
        "y": ["j", "nj", "yh", "nlat","ncl2"],
        "lev": ["deptht", "olevel", "zlev", "olev", "depth","depthu","depthv","ncl1"],
        "bnds": ["bnds", "axis_nbounds", "d2"],
        "vertex": ["vertex", "nvertex", "vertices"],
        # coordinate labels
        "lon": ["longitude", "nav_lon"],
        "lat": ["latitude", "nav_lat"],
        "lev_bounds": [
            "deptht_bounds",
            "lev_bnds",
            "olevel_bounds",
            "zlev_bnds",
        ],
        "lon_bounds": [
            "bounds_lon",
            "bounds_nav_lon",
            "lon_bnds",
            "x_bnds",
            "vertices_longitude",
        ],
        "lat_bounds": [
            "bounds_lat",
            "bounds_nav_lat",
            "lat_bnds",
            "y_bnds",
            "vertices_latitude",
        ],
        "time_bounds": ["time_bnds"],
        # variables
        "uo": ["vozocrtx","siu","u"],
        "vo": ["vomecrty","siv","v"],
        "thetao": ["votemper"],
        "so": ["vosaline"],
        "thkcello": ["e3t","e3u","e3v","e3t_0","e3v_0","e3u_0","e3t_0_field","e3u_0_field","e3v_0_field"],
        "dxv":["e1v"],
        "dyu":["e2u"],
        "sithick": ["sithick", "sit", "SIT", "iicethic"],
        "siconc": ["siconc", "sic", "SIC", "iiceconc"]
    }
    return rename_dict


def lon_to_180(ds):
    lon = ds["lon"].where(ds["lon"] <= 180, ds["lon"] - 360)
    ds = ds.assign_coords(lon=lon)
    return ds

def corr_updown(ds):
    if np.sign(ds.lat[-1,0].values) == -1:
        ds['y']=np.arange(len(ds.y)-1,-1,-1)
        ds=ds.sortby('y')
    return ds

def corr_indices(ds):
    if np.sign(ds.y.min()) == -1:
        ds['y']=np.arange(1,len(ds.y)+1)
    if np.sign(ds.x.min()) == -1:
        ds['x']=np.arange(1,len(ds.x)+1)
    if ds.x.min() <= 0.9:
        ds['x']=np.arange(1,len(ds.x)+1)
    if ds.y.min() <= 0.9:
        ds['y']=np.arange(1,len(ds.y)+1)
    return ds

def corr_CMCC(ds):
    ds=ds.rename({'x':'y','y':'x'})
    return ds


def corr_xyseries(ds):
    try:
        if [i for i in ds[str([i for i in ds.data_vars][0])].dims].index("x") < [i for i in ds[str([i for i in ds.data_vars][0])].dims].index("y"):
            ds=ds.rename({'x':'y','y':'x'})
    except ValueError:
        pass
    return ds

def corr_dims_latlon(ds):
    if ds.lon.dims[0]=='x' and ds.lon.dims[1]=='y':
        print('correcting lat/lon dimensions')
        ds['lon']=ds.lon.transpose("y", "x")
        ds['lat']=ds.lat.transpose("y", "x")
    return ds

def corr_xy_points(ds):
    ds['x']=np.arange(0,len(ds.x))
    ds['y']=np.arange(0,len(ds.y))
    return ds

def corr_lonlat_dims(ds):
    if len(ds.lon.dims) > 2:
        ds['lon']=ds.lon[0]
        ds['lat']=ds.lat[0]
    return ds

def rename_time_lev(ds):
    rename_dict = {
        "lev": ["deptht", "olevel", "zlev", "olev", "depth","depthu","depthv","ncl1"],
        "time": ["time_counter"],
    }
    for i in ds.coords:
        for target, candidates in rename_dict.items():
            if i in candidates:
                ds = ds.rename({i: target})
    return ds

def drop_cyclic_points(ds):
    tt,ind=np.unique(ds.lon,axis=1,return_index=True)
    ds=ds.sel(x=np.sort(ind))
    return ds

def drop_northsouth_duplicate_points(ds):
    tt,ind=np.unique(ds.lat,axis=0,return_index=True)
    ds=ds.sel(y=np.sort(ind))
    return ds

def drop_cyclic_points(ds):
    tt,ind=np.unique(ds.lon,axis=1,return_index=True)
    ds=ds.sel(x=np.sort(ind))
    return ds

def wrapper(ds):
    ds = ds.copy()
    ds = rename_time_lev(ds)
    ds = rename_cmip6(ds,rename_dict=renaming_dict_exp())
    ds = promote_empty_dims(ds)
    ds = broadcast_lonlat(ds)
    ds = correct_coordinates(ds)
    ds = lon_to_180(ds)
    ds = corr_xyseries(ds)
    ds = corr_indices(ds)
    ds = corr_dims_latlon(ds)
    ds = corr_lonlat_dims(ds)
    ds = corr_updown(ds)
    ds = corr_xy_points(ds)
    ds = drop_cyclic_points(ds)
    ds = drop_northsouth_duplicate_points(ds)
    return ds

def distance(lat1,lon1,lat2,lon2):
    '''
    This function calculates the distance between two lat/lon points
    args:
        lat1: latitude of point 1
        lon1: longitude of point 1
        lat2: latitude of point 2
        lon2: longitude of point 2
    '''
    lat11 = np.radians(lat1)
    lon11 = np.radians(lon1)
    lat22 = np.radians(lat2)
    lon22 = np.radians(lon2)
    a = 6378.388    
    dx=np.cos(lat22)*np.cos(lon22)-np.cos(lat11)*np.cos(lon11)
    dy=np.cos(lat22)*np.sin(lon22)-np.cos(lat11)*np.sin(lon11)
    dz=np.sin(lat22)-np.sin(lat11)
    c=np.sqrt(dx*dx+dy*dy+dz*dz)
    distance=a*2*np.arcsin(c/2) 
    return distance

def calc_dxdy(model,u,v,path_mesh):

    dy=xa.DataArray(data=np.zeros(u.lat.shape),coords=u.lat.coords,dims=u.lat.dims)
    dx=xa.DataArray(data=np.zeros(v.lat.shape),coords=v.lat.coords,dims=v.lat.dims)
    for i in tqdm(range(0,len(u.y)-1)):
        #print(i)
        dy[i+1,:]=distance(u.lat[i,:],u.lon[i,:],u.lat[i+1,:],u.lon[i+1,:])
    for i in tqdm(range(0,len(v.x)-1)):
        #print(i)
        dx[:,i+1]=distance(v.lat[:,i],v.lon[:,i],v.lat[:,i+1],v.lon[:,i+1])
            
    mu=(dy*1000).to_dataset(name='dyu')
    mv=(dx*1000).to_dataset(name='dxv')
    mu.to_netcdf(path_mesh+'mesh_dyu_'+model+'.nc')
    mv.to_netcdf(path_mesh+'mesh_dxv_'+model+'.nc')
    return mu,mv

def unique_rows(a):
    a = np.ascontiguousarray(a)
    unique_a = np.unique(a.view([('',a.dtype)]*a.shape[1]))
    return unique_a.view(a.dtype).reshape(unique_a.shape[0],a.shape[1])

def select_area(ds,out_u,out_v):
    min_x=np.nanmin((min(out_u[:,0],default=np.nan),min(out_v[:,0],default=np.nan)))
    max_x=np.nanmax((max(out_u[:,0],default=np.nan),max(out_v[:,0],default=np.nan)))
    min_y=np.nanmin((min(out_u[:,1],default=np.nan),min(out_v[:,1],default=np.nan)))
    max_y=np.nanmax((max(out_u[:,1],default=np.nan),max(out_v[:,1],default=np.nan)))
    ds = ds.sel(x=slice(int(min_x)-1,int(max_x)+1),y=slice(int(min_y)-1,int(max_y)+1))
    return ds

## for indices selection
def _preprocess1(x):
    x=wrapper(x)
    return x.isel(lev=0)

## for indices selection ice
def _preprocess1i(x):
    x=wrapper(x)
    return x

## for actuall fields
def _preprocess2(x, lon_bnds, lat_bnds):
    x=wrapper(x)
    return x.sel(x=slice(*lon_bnds), y=slice(*lat_bnds))


def kugel_2_kart(lat,lon):
    '''
    This function transforms from spherical to kartesian coordinates
    args:
        lat and lon
    '''
    radius =  6378388. # m
    x = radius * np.cos(np.radians(lat)) * np.cos(np.radians(lon))
    y = radius * np.cos(np.radians(lat)) * np.sin(np.radians(lon))
    z = radius * np.sin(np.radians(lat))
    return x,y,z

def kart_2_kugel(x,y,z):
    '''
    This function transforms from kartesian to spherical coordinates
    args:
        x,y,z
    '''
    lat = np.degrees(np.arcsin(z/(np.sqrt(x*x+y*y+z*z))))
    lon2 = []
    for i in range(len(x)):
        if x[i] > 0:
            lon = np.degrees(np.arctan(y[i]/x[i]))
        elif y[i] > 0:
            lon = np.degrees(np.arctan(y[i]/x[i])) + 180
        else:
            lon = np.degrees(np.arctan(y[i]/x[i])) - 180
        lon2=np.append(lon2,lon)
    return lat,lon2






import xarray as xa 
import pandas as pd
import numpy as np
from tqdm import tqdm
try:
    import matplotlib.pyplot as plt
except ImportError:
    print('skipping matplotlib')
import sys
from functools import partial
import time
import xesmf as xe
try:
    from dask.diagnostics import ProgressBar
except ImportError:
    print('skipping dask import')

from xmip.preprocessing import rename_cmip6, promote_empty_dims, broadcast_lonlat, correct_coordinates
import StraitFlux.preprocessing as prepro
import StraitFlux.functions as func
import StraitFlux.functions_VP as func2
from StraitFlux.indices import check_availability_indices, prepare_indices

def vel_projection(strait,model,time_start,time_end,file_u,file_v,file_t,file_z, file_zu=None, file_zv=None,mesh_dxv=0, mesh_dyu=0,coords=0,set_latlon=False,lat_p=0,lon_p=0,path_save='',path_indices='',path_mesh='',Arakawa='',saving=True):
    '''
    This function calculates the u/v crossection by projecting the vectors onto the strait 
    by multiplying with the amount of the vector going through the strait and
    regridding the values onto the T-proj points on the reference line 

    INPUT Parameters:
    strait (str): desired oceanic strait, either pre-defined from indices file or new
    model (str): desired CMIP6 model or reanalysis
    time_start (str or int): starting year
    time_end (str or int): ending year
    file_u (str): path + filename(s) of u field(s); use ice velocities (ui) for ice transports; (multiple files possible, use *; must be possible to combine files over time coordinate)
    file_v (str): path + filename(s) of v field(s); use ice velocities (vi) for ice transports; (multiple files possible, use *)
    file_t (str): path + filename(s) of temperature field(s); (multiple files possible, use *)
    file_z (str): path + filename(s) of cell thickness field(s); (multiple files possible, use *)

    OPTIONAL:
    coords (tuple): coordinates for strait, if not pre-defined: (latitude_start,longitude_start,latitude_end,longitude_end)
    set_latlon: set True if you wish to pass arrays of latitudes and longitudes
    lon (array): longitude coordinates for strait, if not pre-defined. (range -180 to 180; same length as lat needed!)
    lat (array): latitude coordinates for strait, if not pre-defined. (range -90 to 90; same length as lon needed!)
    Arakawa (str): Arakawa-A, Arakawa-B or Arakawa-C; only needed if automatic check fails
    path_save (str): path to save transport data
    path_indices (str): path to save indices data
    path_mesh (str): path to save mesh data

    RETURNS:
    crosssection of currents at given section

    '''
    
    partial_func = partial(prepro._preprocess1)

    try:
        indices=xa.open_dataset(path_indices+model+'_'+strait+'_indices.nc')
    except OSError:
        print('calc indices')
        print('read and load files for indices')
        ti = xa.open_mfdataset(file_t, preprocess=partial_func).isel(time=0)
        ui = xa.open_mfdataset(file_u, preprocess=partial_func).isel(time=0)
        vi = xa.open_mfdataset(file_v, preprocess=partial_func).isel(time=0)
        with ProgressBar():
            ti=ti.load()
            ui=ui.load()
            vi=vi.load()
        indices,line = check_availability_indices(ti,strait,model,coords,lon_p,lat_p,set_latlon)
        i2=indices.indices.where(indices.indices!=0) 
        try:
            (ti.thetao/ti.thetao).plot(add_colorbar=False)
            plt.scatter(i2[:,2],i2[:,3],color='tab:red',s=0.1,marker='x')
            plt.scatter(i2[:,0],i2[:,1],color='tab:red',s=0.1,marker='x')
            plt.title(model+'_'+strait,fontsize=14)
            plt.ylabel('y',fontsize=14)
            plt.xlabel('x',fontsize=14)
            plt.savefig(path_save+strait+'_'+model+'_indices.png')
            plt.close()
        except NameError:
            print('skipping Plot')
        indices.to_netcdf(path_indices+model+'_'+strait+'_indices.nc')

    #######
    if Arakawa in ['Arakawa-A','Arakawa-B','Arakawa-C']:
        grid=Arakawa
    elif Arakawa == '':
        try:
            file = open(path_mesh+model+'grid.txt', 'r')
            grid= file.read()
        except OSError:
            partial_func = partial(prepro._preprocess1)
            ti = xa.open_mfdataset(file_t, preprocess=partial_func).isel(time=0)
            ui = xa.open_mfdataset(file_u, preprocess=partial_func).isel(time=0)
            vi = xa.open_mfdataset(file_v, preprocess=partial_func).isel(time=0)
            try:
                with ProgressBar():
                    ti=ti.load()
                    ui=ui.load()
                    vi=vi.load()
            except NameError:
                ti=ti.load()
                ui=ui.load()
                vi=vi.load()
            grid = func.check_Arakawa(ui,vi,ti,model)
            with open(path_mesh+model+'grid.txt', 'w') as f:
                f.write(grid)
    else:
        print('grid not known')
        sys.exit()
    ################     
                 

    try:
        indices = xa.open_dataset(path_save+'indices_long_'+model+strait+'.nc')
        betrag_u = np.loadtxt(path_save+'betrag_u_'+model+strait+'.txt')
        betrag_v = np.loadtxt(path_save+'betrag_v_'+model+strait+'.txt')
        tu = np.loadtxt(path_save+'tu_'+model+strait+'.txt')
        tv = np.loadtxt(path_save+'tv_'+model+strait+'.txt')
        un = np.loadtxt(path_save+'un_'+model+strait+'.txt')
        dist_listT_kurz2 = np.loadtxt(path_save+'dx_'+model+strait+'.txt')
        T_proj_points=xa.open_dataset(path_save+'T_proj_points_'+model+strait+'.nc')
        
    except FileNotFoundError:
        try:
            T_data=ti
            u_data=ui
            v_data=vi
        except NameError:      
            T_data = xa.open_mfdataset(file_t, preprocess=partial_func).isel(time=0)
            u_data = xa.open_mfdataset(file_u, preprocess=partial_func).isel(time=0)
            v_data = xa.open_mfdataset(file_v, preprocess=partial_func).isel(time=0)
            with ProgressBar():
                T_data=T_data.load()
                u_data=u_data.load()
                v_data=v_data.load()
        start = time.time()
        indices,ref_line=check_availability_indices(T_data,strait,model,coords,lon_p,lat_p,set_latlon)
        end = time.time()
        print(end - start)
        start = time.time()
        un,u_line,v_line,T_line,u_line2,v_line2=func2.select_indices(indices,u_data,v_data,T_data)
        proj_u,proj_v,T_proj, u_proj, v_proj=func2.proj_vec(ref_line,u_line,v_line,T_line,u_line2,v_line2)
        T_proj_points,dist_listT,dist_listT_kurz2=func2.calc_interpolation_points(indices,T_data, ref_line)
        betrag_u=func2.calc_betrag(proj_u)
        betrag_v=func2.calc_betrag(proj_v)
        tu,tv=func2.multi_factors(T_line,u_line,v_line,T_proj,u_proj,v_proj)
        end = time.time()
        print(end - start)
        
        if saving == True:
            np.savetxt(path_save+'betrag_u_'+model+strait+'.txt', betrag_u)
            np.savetxt(path_save+'betrag_v_'+model+strait+'.txt', betrag_v)
            np.savetxt(path_save+'tu_'+model+strait+'.txt', tu)
            np.savetxt(path_save+'tv_'+model+strait+'.txt', tv)
            np.savetxt(path_save+'un_'+model+strait+'.txt', un)
            np.savetxt(path_save+'dx_'+model+strait+'.txt', dist_listT_kurz2)
            T_proj_points.to_netcdf(path_save+'T_proj_points_'+model+strait+'.nc')
            indices.to_netcdf(path_save+'indices_long_'+model+strait+'.nc')

    
    out_u,out_v,out_u_vz = prepare_indices(indices)
    min_x=np.nanmin((min(out_u[:,0],default=np.nan),min(out_v[:,0],default=np.nan)))
    max_x=np.nanmax((max(out_u[:,0],default=np.nan),max(out_v[:,0],default=np.nan)))
    min_y=np.nanmin((min(out_u[:,1],default=np.nan),min(out_v[:,1],default=np.nan)))
    max_y=np.nanmax((max(out_u[:,1],default=np.nan),max(out_v[:,1],default=np.nan)))

    if min_x == -1:
        min_x = 0
        max_x = max_x + 1
    #print(min_x,max_x,min_y,max_y)
    #sys.exit()
    print('read t, u and v fields')
    partial_func = partial(prepro._preprocess2,lon_bnds=(int(min_x)-2,int(max_x)+2),lat_bnds=(int(min_y)-2,int(max_y)+2))
    t = xa.open_mfdataset(file_t, preprocess=partial_func,chunks={'time':1}).sel(time=slice(str(time_start),str(time_end)))
    u = xa.open_mfdataset(file_u, preprocess=partial_func,chunks={'time':1}).sel(time=slice(str(time_start),str(time_end)))
    v = xa.open_mfdataset(file_v, preprocess=partial_func,chunks={'time':1}).sel(time=slice(str(time_start),str(time_end)))
    deltaz = xa.open_mfdataset(file_z, preprocess=partial_func,chunks={'time':1})[['thkcello']]
    
    if 'time' in deltaz.dims:
        deltaz=deltaz.sel(time=slice(str(time_start),str(time_end)))

    print('load t, u and v fields')
    with ProgressBar():
        t=t.load()
        u=u.load()
        v=v.load()
        deltaz=deltaz.load()
        
        
    dzu3,dzv3 = func.calc_dz_faces(deltaz,grid,model,path_mesh)
    
    start = time.time()
    print('calculating regridder')
    regridder_u=xe.Regridder(u,T_proj_points,'bilinear',ignore_degenerate=True)
    regridder_v=xe.Regridder(v,T_proj_points,'bilinear',ignore_degenerate=True)
    end = time.time()
    print(end - start)    
    u_beitrag = np.zeros((len(u.time),len(u.lev),len(T_proj_points.lat))) # time,z=75,x=coords of strait
    v_beitrag = np.zeros((len(u.time),len(u.lev),len(T_proj_points.lat)))

    u_trans = np.zeros((len(u.time),len(u.lev),len(u.y),len(u.x)))
    v_trans = np.zeros((len(u.time),len(u.lev),len(u.y),len(u.x)))
    start = time.time()
    for i in range(len(un)):
        if 'time' in deltaz.dims:
            u_trans[:,:,int(un[i,1]-min_y+2),int(un[i,0]-min_x+2)]=u.uo[:,:,int(un[i,1]-min_y+2),int(un[i,0]-min_x+2)]*dzu3[:,:,int(un[i,1]-min_y+2),int(un[i,0]-min_x+2)].values*betrag_u[i]*tu[i]#
            v_trans[:,:,int(un[i,1]-min_y+2),int(un[i,0]-min_x+2)]=v.vo[:,:,int(un[i,1]-min_y+2),int(un[i,0]-min_x+2)]*dzv3[:,:,int(un[i,1]-min_y+2),int(un[i,0]-min_x+2)].values*betrag_v[i]*tv[i]        
        else:
            u_trans[:,:,int(un[i,1]-min_y+2),int(un[i,0]-min_x+2)]=u.uo[:,:,int(un[i,1]-min_y+2),int(un[i,0]-min_x+2)]*dzu3[:,int(un[i,1]-min_y+2),int(un[i,0]-min_x+2)].values*betrag_u[i]*tu[i]#
            v_trans[:,:,int(un[i,1]-min_y+2),int(un[i,0]-min_x+2)]=v.vo[:,:,int(un[i,1]-min_y+2),int(un[i,0]-min_x+2)]*dzv3[:,int(un[i,1]-min_y+2),int(un[i,0]-min_x+2)].values*betrag_v[i]*tv[i]#

    
    
    u['uo'][:,:,:,:]=u_trans[:,:,:,:]#
    v['vo'][:,:,:,:]=v_trans[:,:,:,:]#
    end = time.time()
    print(end - start) 
    print('regridding')
    start = time.time()
    u_l=regridder_u(u.uo.fillna(0))#
    v_l=regridder_v(v.vo.fillna(0))#
    end = time.time()
    print(end - start) 
    for m in range(len(u_l.lat)):
        for k in range(len(u.lev)):
            u_beitrag[:,k,m] = u_l[:,k].isel(lat=m,lon=m).values
            v_beitrag[:,k,m] = v_l[:,k].isel(lat=m,lon=m).values
            
    gesamt=np.where(np.isnan(u_beitrag),0,u_beitrag)+np.where(np.isnan(v_beitrag),0,v_beitrag)
    dz2=np.gradient(u.lev)
    dz3=np.transpose([dz2]*np.shape(gesamt)[-1])
    uv_tot = xa.Dataset({'uv':(('time','depth','x'),gesamt/dz3*(u_beitrag[0]/u_beitrag[0])),'dx_int':(('x'),dist_listT_kurz2),'dz_int':(('depth'),dz2.data)},coords=dict(time=u.time,depth=u.lev.data,x=np.cumsum(dist_listT_kurz2)))
    uv_tot.to_netcdf(path_save+strait+'_crosssection_uv_'+model+'_'+str(time_start)+'-'+str(time_end)+'.nc')
    return uv_tot

def uvT_projection(strait,model,time_start,time_end,file_u,file_v,file_t,file_z,coords=0,set_latlon=False,lat_p=0,lon_p=0,path_save='',path_indices='',path_mesh='',Arakawa='',saving=True):
    '''
    This function calculates the u/v crossection by projecting the vectors onto the strait 
    by multiplying with the amount of the vector going through the strait and
    regridding the values onto the T-proj points on the reference line 

    INPUT Parameters:
    strait (str): desired oceanic strait, either pre-defined from indices file or new
    model (str): desired CMIP6 model or reanalysis
    time_start (str or int): starting year
    time_end (str or int): ending year
    file_u (str): path + filename(s) of u field(s); use ice velocities (ui) for ice transports; (multiple files possible, use *; must be possible to combine files over time coordinate)
    file_v (str): path + filename(s) of v field(s); use ice velocities (vi) for ice transports; (multiple files possible, use *)
    file_t (str): path + filename(s) of temperature field(s); (multiple files possible, use *)
    file_z (str): path + filename(s) of cell thickness field(s); (multiple files possible, use *)

    OPTIONAL:
    coords (tuple): coordinates for strait, if not pre-defined: (latitude_start,longitude_start,latitude_end,longitude_end)
    set_latlon: set True if you wish to pass arrays of latitudes and longitudes
    lon (array): longitude coordinates for strait, if not pre-defined. (range -180 to 180; same length as lat needed!)
    lat (array): latitude coordinates for strait, if not pre-defined. (range -90 to 90; same length as lon needed!)
    Arakawa (str): Arakawa-A, Arakawa-B or Arakawa-C; only needed if automatic check fails
    path_save (str): path to save transport data
    path_indices (str): path to save indices data
    path_mesh (str): path to save mesh data

    RETURNS:
    crosssection of currents at given section

    '''
    
    partial_func = partial(prepro._preprocess1)

    try:
        indices=xa.open_dataset(path_indices+model+'_'+strait+'_indices.nc')
    except OSError:
        print('calc indices')
        print('read and load files for indices')
        ti = xa.open_mfdataset(file_t, preprocess=partial_func).isel(time=0)
        ui = xa.open_mfdataset(file_u, preprocess=partial_func).isel(time=0)
        vi = xa.open_mfdataset(file_v, preprocess=partial_func).isel(time=0)
        with ProgressBar():
            ti=ti.load()
            ui=ui.load()
            vi=vi.load()
        indices,line = check_availability_indices(ti,strait,model,coords,lon_p,lat_p,set_latlon)
        i2=indices.indices.where(indices.indices!=0) 
        try:
            (ti.thetao/ti.thetao).plot(add_colorbar=False)
            plt.scatter(i2[:,2],i2[:,3],color='tab:red',s=0.1,marker='x')
            plt.scatter(i2[:,0],i2[:,1],color='tab:red',s=0.1,marker='x')
            plt.title(model+'_'+strait,fontsize=14)
            plt.ylabel('y',fontsize=14)
            plt.xlabel('x',fontsize=14)
            plt.savefig(path_save+strait+'_'+model+'_indices.png')
            plt.close()
        except NameError:
            print('skipping Plot')
        indices.to_netcdf(path_indices+model+'_'+strait+'_indices.nc')

    #######
    if Arakawa in ['Arakawa-A','Arakawa-B','Arakawa-C']:
        grid=Arakawa
    elif Arakawa == '':
        try:
            file = open(path_mesh+model+'grid.txt', 'r')
            grid= file.read()
        except OSError:
            partial_func = partial(prepro._preprocess1)
            ti = xa.open_mfdataset(file_t, preprocess=partial_func).isel(time=0)
            ui = xa.open_mfdataset(file_u, preprocess=partial_func).isel(time=0)
            vi = xa.open_mfdataset(file_v, preprocess=partial_func).isel(time=0)
            try:
                with ProgressBar():
                    ti=ti.load()
                    ui=ui.load()
                    vi=vi.load()
            except NameError:
                ti=ti.load()
                ui=ui.load()
                vi=vi.load()
            grid = func.check_Arakawa(ui,vi,ti,model)
            with open(path_mesh+model+'grid.txt', 'w') as f:
                f.write(grid)
    else:
        print('grid not known')
        sys.exit()
    ################     
                 

    try:
        indices = xa.open_dataset(path_save+'indices_long_'+model+strait+'.nc')
        betrag_u = np.loadtxt(path_save+'betrag_u_'+model+strait+'.txt')
        betrag_v = np.loadtxt(path_save+'betrag_v_'+model+strait+'.txt')
        tu = np.loadtxt(path_save+'tu_'+model+strait+'.txt')
        tv = np.loadtxt(path_save+'tv_'+model+strait+'.txt')
        un = np.loadtxt(path_save+'un_'+model+strait+'.txt')
        dist_listT_kurz2 = np.loadtxt(path_save+'dx_'+model+strait+'.txt')
        T_proj_points=xa.open_dataset(path_save+'T_proj_points_'+model+strait+'.nc')
        
    except FileNotFoundError:
        try:
            T_data=ti
            u_data=ui
            v_data=vi
        except NameError:      
            T_data = xa.open_mfdataset(file_t, preprocess=partial_func).isel(time=0)
            u_data = xa.open_mfdataset(file_u, preprocess=partial_func).isel(time=0)
            v_data = xa.open_mfdataset(file_v, preprocess=partial_func).isel(time=0)
            with ProgressBar():
                T_data=T_data.load()
                u_data=u_data.load()
                v_data=v_data.load()
        start = time.time()
        indices,ref_line=check_availability_indices(T_data,strait,model,coords,lon_p,lat_p,set_latlon)
        end = time.time()
        print(end - start)
        start = time.time()
        un,u_line,v_line,T_line,u_line2,v_line2=func2.select_indices(indices,u_data,v_data,T_data)
        proj_u,proj_v,T_proj, u_proj, v_proj=func2.proj_vec(ref_line,u_line,v_line,T_line,u_line2,v_line2)
        T_proj_points,dist_listT,dist_listT_kurz2=func2.calc_interpolation_points(indices,T_data, ref_line)
        betrag_u=func2.calc_betrag(proj_u)
        betrag_v=func2.calc_betrag(proj_v)
        tu,tv=func2.multi_factors(T_line,u_line,v_line,T_proj,u_proj,v_proj)
        end = time.time()
        print(end - start)
        
        if saving == True:
            np.savetxt(path_save+'betrag_u_'+model+strait+'.txt', betrag_u)
            np.savetxt(path_save+'betrag_v_'+model+strait+'.txt', betrag_v)
            np.savetxt(path_save+'tu_'+model+strait+'.txt', tu)
            np.savetxt(path_save+'tv_'+model+strait+'.txt', tv)
            np.savetxt(path_save+'un_'+model+strait+'.txt', un)
            np.savetxt(path_save+'dx_'+model+strait+'.txt', dist_listT_kurz2)
            T_proj_points.to_netcdf(path_save+'T_proj_points_'+model+strait+'.nc')
            indices.to_netcdf(path_save+'indices_long_'+model+strait+'.nc')

    
    out_u,out_v,out_u_vz = prepare_indices(indices)
    min_x=np.nanmin((min(out_u[:,0],default=np.nan),min(out_v[:,0],default=np.nan)))
    max_x=np.nanmax((max(out_u[:,0],default=np.nan),max(out_v[:,0],default=np.nan)))
    min_y=np.nanmin((min(out_u[:,1],default=np.nan),min(out_v[:,1],default=np.nan)))
    max_y=np.nanmax((max(out_u[:,1],default=np.nan),max(out_v[:,1],default=np.nan)))

    if min_x == -1:
        min_x = 0
        max_x = max_x + 1
    #print(min_x,max_x,min_y,max_y)
    #sys.exit()
    print('read t, u and v fields')
    partial_func = partial(prepro._preprocess2,lon_bnds=(int(min_x)-2,int(max_x)+2),lat_bnds=(int(min_y)-2,int(max_y)+2))
    t = xa.open_mfdataset(file_t, preprocess=partial_func,chunks={'time':1}).sel(time=slice(str(time_start),str(time_end)))
    u = xa.open_mfdataset(file_u, preprocess=partial_func,chunks={'time':1}).sel(time=slice(str(time_start),str(time_end)))
    v = xa.open_mfdataset(file_v, preprocess=partial_func,chunks={'time':1}).sel(time=slice(str(time_start),str(time_end)))
    deltaz = xa.open_mfdataset(file_z, preprocess=partial_func,chunks={'time':1})[['thkcello']]
    
    if 'time' in deltaz.dims:
        deltaz=deltaz.sel(time=slice(str(time_start),str(time_end)))

    print('load t, u and v fields')
    with ProgressBar():
        t=t.load()
        u=u.load()
        v=v.load()
        deltaz=deltaz.load()
        
        
    dzu3,dzv3 = func.calc_dz_faces(deltaz,grid,model,path_mesh)
    
    start = time.time()
    print('calculating regridder')
    regridder_u=xe.Regridder(u,T_proj_points,'bilinear',ignore_degenerate=True)
    regridder_v=xe.Regridder(v,T_proj_points,'bilinear',ignore_degenerate=True)
    end = time.time()
    print(end - start)    
    u_beitrag = np.zeros((len(u.time),len(u.lev),len(T_proj_points.lat))) # time,z=75,x=coords of strait
    v_beitrag = np.zeros((len(u.time),len(u.lev),len(T_proj_points.lat)))

    u_trans = np.zeros((len(u.time),len(u.lev),len(u.y),len(u.x)))
    v_trans = np.zeros((len(u.time),len(u.lev),len(u.y),len(u.x)))
    start = time.time()

    thetau=t.thetao.rolling(x=2,min_periods=1).mean()
    thetav=t.thetao.rolling(y=2,min_periods=1).mean()

    for i in range(len(un)):
        if 'time' in deltaz.dims:
            u_trans[:,:,int(un[i,1]-min_y+2),int(un[i,0]-min_x+2)]=u.uo[:,:,int(un[i,1]-min_y+2),int(un[i,0]-min_x+2)]*thetau[:,:,int(un[i,1]-min_y+2),int(un[i,0]-min_x+2)]*dzu3[:,:,int(un[i,1]-min_y+2),int(un[i,0]-min_x+2)].values*betrag_u[i]*tu[i]#
            v_trans[:,:,int(un[i,1]-min_y+2),int(un[i,0]-min_x+2)]=v.vo[:,:,int(un[i,1]-min_y+2),int(un[i,0]-min_x+2)]*thetav[:,:,int(un[i,1]-min_y+2),int(un[i,0]-min_x+2)]*dzv3[:,:,int(un[i,1]-min_y+2),int(un[i,0]-min_x+2)].values*betrag_v[i]*tv[i]        
        else:
            u_trans[:,:,int(un[i,1]-min_y+2),int(un[i,0]-min_x+2)]=u.uo[:,:,int(un[i,1]-min_y+2),int(un[i,0]-min_x+2)]*thetau[:,:,int(un[i,1]-min_y+2),int(un[i,0]-min_x+2)]*dzu3[:,int(un[i,1]-min_y+2),int(un[i,0]-min_x+2)].values*betrag_u[i]*tu[i]#
            v_trans[:,:,int(un[i,1]-min_y+2),int(un[i,0]-min_x+2)]=v.vo[:,:,int(un[i,1]-min_y+2),int(un[i,0]-min_x+2)]*thetav[:,:,int(un[i,1]-min_y+2),int(un[i,0]-min_x+2)]*dzv3[:,int(un[i,1]-min_y+2),int(un[i,0]-min_x+2)].values*betrag_v[i]*tv[i]#

    
    
    u['uo'][:,:,:,:]=u_trans[:,:,:,:]#
    v['vo'][:,:,:,:]=v_trans[:,:,:,:]#
    end = time.time()
    print(end - start) 
    print('regridding')
    start = time.time()
    u_l=regridder_u(u.uo.fillna(0))#
    v_l=regridder_v(v.vo.fillna(0))#
    end = time.time()
    print(end - start) 
    for m in range(len(u_l.lat)):
        for k in range(len(u.lev)):
            u_beitrag[:,k,m] = u_l[:,k].isel(lat=m,lon=m).values
            v_beitrag[:,k,m] = v_l[:,k].isel(lat=m,lon=m).values
            
    gesamt=np.where(np.isnan(u_beitrag),0,u_beitrag)+np.where(np.isnan(v_beitrag),0,v_beitrag)
    dz2=np.gradient(u.lev)
    dz3=np.transpose([dz2]*np.shape(gesamt)[-1])
    uv_tot = xa.Dataset({'uvT':(('time','depth','x'),gesamt/dz3*(u_beitrag[0]/u_beitrag[0])),'dx_int':(('x'),dist_listT_kurz2),'dz_int':(('depth'),dz2.data)},coords=dict(time=u.time,depth=u.lev.data,x=np.cumsum(dist_listT_kurz2)))
    uv_tot.to_netcdf(path_save+strait+'_crosssection_uv_'+model+'_'+str(time_start)+'-'+str(time_end)+'.nc')
    return uv_tot

def TS_interp(product,strait,model,time_start,time_end,file_u,file_t,file_s='',coords=0,set_latlon=False,lat_p=0,lon_p=0,path_save='',path_indices='',path_mesh='',saving=True):
    '''
    This function calculates the u/v crossection by projecting the vectors onto the strait 
    by multiplying with the amount of the vector going through the strait and
    regridding the values onto the T-proj points on the reference line 

    INPUT Parameters:
    product (str): T or S
    strait (str): desired oceanic strait, either pre-defined from indices file or new
    model (str): desired CMIP6 model or reanalysis
    time_start (str or int): starting year
    time_end (str or int): ending year
    file_u (str): path + filename(s) of u field(s); use ice velocities (ui) for ice transports; (multiple files possible, use *; must be possible to combine files over time coordinate)
    file_t (str): path + filename(s) of temperature field(s); (multiple files possible, use *)

    OPTIONAL:
    coords (tuple): coordinates for strait, if not pre-defined: (latitude_start,longitude_start,latitude_end,longitude_end)
    set_latlon: set True if you wish to pass arrays of latitudes and longitudes
    lon (array): longitude coordinates for strait, if not pre-defined. (range -180 to 180; same length as lat needed!)
    lat (array): latitude coordinates for strait, if not pre-defined. (range -90 to 90; same length as lon needed!)
    file_s (str): only needed for salinity transports; path + filename(s) of salinity field(s); (multiple files possible, use *)
    Arakawa (str): Arakawa-A, Arakawa-B or Arakawa-C; only needed if automatic check fails
    path_save (str): path to save transport data
    path_indices (str): path to save indices data
    path_mesh (str): path to save mesh data

    RETURNS:
    crosssection of temperature (T) or salinity (S) at given section

    '''
    
    partial_func = partial(prepro._preprocess1)
    try:
        indices=xa.open_dataset(path_indices+model+'_'+strait+'_indices.nc')
    except OSError:
        print('calc indices')
        print('read and load files for indices')
        ti = xa.open_mfdataset(file_t, preprocess=partial_func).isel(time=0)
        with ProgressBar():
            ti=ti.load()
        indices,line = check_availability_indices(ti,strait,model,coords,lon_p,lat_p,set_latlon)
        i2=indices.indices.where(indices.indices!=0) 
        try:
            (ti.thetao/ti.thetao).plot(add_colorbar=False)
            plt.scatter(i2[:,2],i2[:,3],color='tab:red',s=0.1,marker='x')
            plt.scatter(i2[:,0],i2[:,1],color='tab:red',s=0.1,marker='x')
            plt.title(model+'_'+strait,fontsize=14)
            plt.ylabel('y',fontsize=14)
            plt.xlabel('x',fontsize=14)
            plt.savefig(path_save+strait+'_'+model+'_indices.png')
            plt.close()
        except NameError:
            print('skipping Plot')
        indices.to_netcdf(path_indices+model+'_'+strait+'_indices.nc')
        
    out_u,out_v,out_u_vz = prepare_indices(indices)
    min_x=np.nanmin((min(out_u[:,0],default=np.nan),min(out_v[:,0],default=np.nan)))
    max_x=np.nanmax((max(out_u[:,0],default=np.nan),max(out_v[:,0],default=np.nan)))
    min_y=np.nanmin((min(out_u[:,1],default=np.nan),min(out_v[:,1],default=np.nan)))
    max_y=np.nanmax((max(out_u[:,1],default=np.nan),max(out_v[:,1],default=np.nan)))

    if min_x == -1:
        min_x = 0
        max_x = max_x + 1
    #print(min_x,max_x,min_y,max_y)
    
    
    try:
        T_proj_points=xa.open_dataset(path_save+'T_proj_points_'+model+strait+'.nc')
        dist_listT_kurz2 = np.loadtxt(path_save+'dx_'+model+strait+'.txt')
    except FileNotFoundError:
        try:
            T_data=ti
        except NameError:      
            T_data = xa.open_mfdataset(file_t, preprocess=partial_func).isel(time=0)
            with ProgressBar():
                T_data=T_data.load()
        start = time.time()
        indices,ref_line=check_availability_indices(T_data,strait,model,coords,lon_p,lat_p,set_latlon)
        T_proj_points,dist_listT,dist_listT_kurz2=func2.calc_interpolation_points(indices,T_data, ref_line)
        
        if saving == True:
            np.savetxt(path_save+'dx_'+model+strait+'.txt', dist_listT_kurz2)
            T_proj_points.to_netcdf(path_save+'T_proj_points_'+model+strait+'.nc')
    

    print('read t and/or s fields')
    partial_func = partial(prepro._preprocess2,lon_bnds=(int(min_x)-5,int(max_x)+5),lat_bnds=(int(min_y)-5,int(max_y)+5))
    if product == 'T':
        t = xa.open_mfdataset(file_t, preprocess=partial_func,chunks={'time':1}).sel(time=slice(str(time_start),str(time_end)))
    elif product == 'S':
        t = xa.open_mfdataset(file_s, preprocess=partial_func,chunks={'time':1}).sel(time=slice(str(time_start),str(time_end)))

    u = xa.open_mfdataset(file_u, preprocess=partial_func,chunks={'time':1}).sel(time=slice(str(time_start),str(time_end))).isel(time=0)
    with ProgressBar():
        t=t.load()
        u=u.load() 
    if product == 'T':
        t['mask'] = t.thetao[0]/t.thetao[0]
        t['mask'] = xa.where(~np.isnan(t['mask']),1,0)
    elif product == 'S':
        t['mask'] = t.so[0]/t.so[0]
        t['mask'] = xa.where(~np.isnan(t['mask']),1,0)
    
    regridder=[]
    print('calculating regridder')
    for s in tqdm(range(len(t.lev))):
        regridder_T=xe.Regridder(t.isel(lev=s),T_proj_points,'bilinear',ignore_degenerate=True,extrap_method='nearest_s2d')
        regridder=np.append(regridder,regridder_T)
    
    T_beitrag2 = np.zeros((len(t.time),len(t.lev),len(T_proj_points.lat),len(T_proj_points.lat)))
    
    print('regridding')
    if product == 'T':
        for s in range(len(t.lev)):
            T_beitrag2[:,s,:,:] = regridder[s](t['thetao'].isel(lev=s))
    elif product == 'S':
        for s in range(len(t.lev)):
            T_beitrag2[:,s,:,:] = regridder[s](t['so'].isel(lev=s))
        
    T_beitrag = np.zeros((len(t.time),len(t.lev),len(T_proj_points.lat)))
    
    for j in range(len(T_proj_points.lat)):
        for k in range(len(t.lev)): ##75
            T_beitrag[:,k,j] = T_beitrag2[:,k,j,j]

            
    ## Mask same as for uv profiles:
    M_beitrag = np.zeros((len(t.lev),len(T_proj_points.lat)))
    regridder_M=xe.Regridder(u,T_proj_points,'bilinear',ignore_degenerate=True)
    M=regridder_M(u.uo.fillna(0))
    for m in range(len(M.lat)):
        for k in range(len(M.lev)):
            M_beitrag[k,m] = M[k].isel(lat=m,lon=m).values
    if product == 'T':            
        T_tot = xa.Dataset({'T':(('time','depth','x'),T_beitrag*(M_beitrag/M_beitrag))},coords=dict(time=t.time,depth=t.lev.data,x=np.cumsum(dist_listT_kurz2)))
        T_tot.to_netcdf(path_save+strait+'_crosssection_T_'+model+'_'+str(time_start)+'-'+str(time_end)+'.nc')
    elif product == 'S':            
        T_tot = xa.Dataset({'S':(('time','depth','x'),T_beitrag*(M_beitrag/M_beitrag))},coords=dict(time=t.time,depth=t.lev.data,x=np.cumsum(dist_listT_kurz2)))
        T_tot.to_netcdf(path_save+strait+'_crosssection_S_'+model+'_'+str(time_start)+'-'+str(time_end)+'.nc')
    return T_tot



def calc_MOC(strait,model,time_start,time_end,file_u,file_v,file_t,file_z,coords=0,set_latlon=False,lat_p=0,lon_p=0,path_save='',path_indices='',path_mesh='',Arakawa='',saving=True):
    uv=vel_projection(strait,model,time_start,time_end,file_u,file_v,file_t,file_z,coords=0,set_latlon=False,lat_p=0,lon_p=0,path_save='',path_indices='',path_mesh='',Arakawa='',saving=True)
    MOC=(uv.uv*uv.dx_int*uv.dz_int).sum(axis=2)[:,::-1].cumsum('depth')/-1
    return MOC



from setuptools import setup

setup(
    name = 'StraitFlux',
    packages = ['StraitFlux'],
    version = '1.1.0',
    install_requires=['xarray', 'netcdf4', 'xesmf', 'xmip', 'tqdm', 'matplotlib', 'dask'], #external packages as dependencies
    python_requires=">=3.10",
    description = 'StraitFlux',
    long_description = 'StraitFlux - calculates oceanic transports and crosssections on different model grids.',
    author = 'Susanna Winkelbauer',
    author_email = 'susanna.winkelbauer@univie.ac.at',
    url = 'https://github.com/susannawinkelbauer/StraitFlux',
    download_url = 'https://github.com/susannawinkelbauer/StraitFlux/archive/refs/tags/v1.1.0.tar.gz',
    keywords = ['python', 'oceanic transports','curvilinear grids'],
    classifiers = ["Intended Audience :: Science/Research", "Programming Language :: Python :: 3",],
)



import xarray as xa
import numpy as np
try:
    import matplotlib.pyplot as plt
except ImportError:
    print('skipping matplotlib')
from tqdm import tqdm
import sys
from dask.diagnostics import ProgressBar
from StraitFlux.indices import check_availability_indices, prepare_indices
import StraitFlux.preprocessing as prepro

def select_indices(indices,u_data,v_data,T_data):
    '''
    This function selects indices on T,u and v grid
    args:
        path_indices: path with indices files
        strait: Fram,Barents,GSR,etc.
        path_raw_data: path to raw fata files
        reana: ORAS5,FOAM,CGLORS,GLORYS2V4
    '''

    selection = indices.indices.values
    sels = np.zeros((len(selection),2))
    i=0
    for i in range(len(selection)):
        if selection[i,0] == 0 and selection[i,1] == 0:
            sels[i,0] = selection[i,2]
            sels[i,1] = selection[i,3]
        else:
            sels[i,0] = selection[i,0]
            sels[i,1] = selection[i,1]

    ### take points north, east, west and south of indices
    uniques=np.unique(sels,axis=0)
    uniques_n=uniques.copy()
    uniques_n[:,0]=uniques[:,0]+1
    uniques_e=uniques.copy()
    uniques_e[:,1]=uniques[:,1]+1
    uniques_s=uniques.copy()
    uniques_s[:,0]=uniques[:,0]-1
    uniques_w=uniques.copy()
    uniques_w[:,1]=uniques[:,1]-1
    uniques_ne=uniques.copy()
    uniques_ne[:,:]=uniques+1
    uniques_se=uniques.copy()
    uniques_se[:,1]=uniques[:,1]+1
    uniques_se[:,0]=uniques[:,0]-1
    uniques_nw=uniques.copy()
    uniques_nw[:,1]=uniques[:,1]-1
    uniques_nw[:,0]=uniques[:,0]+1
    uniques_sw=uniques.copy()
    uniques_sw[:,:]=uniques-1
    
    

    uniques2=np.concatenate((uniques,uniques_n,uniques_e,uniques_s,uniques_w,uniques_ne,uniques_se,uniques_nw,uniques_sw),axis=0) 
    uniques_new=np.unique(uniques2,axis=0)


    u_points_lons, u_points_lats = [], []
    v_points_lons, v_points_lats = [], []
    T_points_lons, T_points_lats = [], []
    u_points_lons2, u_points_lats2 = [], []
    v_points_lons2, v_points_lats2 = [], []
    T_points_lons2, T_points_lats2 = [], []
    
    
    # coordinates of u_points:
    i=0
    for i in range(len(uniques_new)):
        # selected points:
        u_points_lon = u_data.isel(x=int(uniques_new[i,0]),y=int(uniques_new[i,1])).lon.values
        u_points_lons = np.append(u_points_lons,u_points_lon)
        u_points_lat = u_data.isel(x=int(uniques_new[i,0]),y=int(uniques_new[i,1])).lat.values
        u_points_lats = np.append(u_points_lats,u_points_lat)
    # coordinates of v_points:
    j=0
    for j in range(len(uniques_new)):
        v_points_lon = v_data.isel(x=int(uniques_new[j,0]),y=int(uniques_new[j,1])).lon.values
        v_points_lons = np.append(v_points_lons,v_points_lon)
        v_points_lat = v_data.isel(x=int(uniques_new[j,0]),y=int(uniques_new[j,1])).lat.values
        v_points_lats = np.append(v_points_lats,v_points_lat)
    # coordinates of T_points:
    k=0
    for k in range(len(uniques_new)):
        T_points_lon = T_data.isel(x=int(uniques_new[k,0]),y=int(uniques_new[k,1])).lon.values
        T_points_lons = np.append(T_points_lons,T_points_lon)
        T_points_lat = T_data.isel(x=int(uniques_new[k,0]),y=int(uniques_new[k,1])).lat.values
        T_points_lats = np.append(T_points_lats,T_points_lat)
    k=0
    for k in range(len(sels)):
        T_points_lon2 = T_data.isel(x=int(sels[k,0]),y=int(sels[k,1])).lon.values
        T_points_lons2 = np.append(T_points_lons2,T_points_lon2)
        T_points_lat2 = T_data.isel(x=int(sels[k,0]),y=int(sels[k,1])).lat.values
        T_points_lats2 = np.append(T_points_lats2,T_points_lat2)

    i=0
    for i in range(len(uniques_new)):
        # selected points:
        try:
            u_points_lon2 = u_data.isel(x=int(uniques_new[i,0]+1),y=int(uniques_new[i,1])).lon.values
        except IndexError:
            u_points_lon2 = u_data.isel(x=0,y=int(uniques_new[i,1])).lon.values
        u_points_lons2 = np.append(u_points_lons2,u_points_lon2)
        try:
            u_points_lat2 = u_data.isel(x=int(uniques_new[i,0]+1),y=int(uniques_new[i,1])).lat.values
        except IndexError:
            u_points_lat2 = u_data.isel(x=0,y=int(uniques_new[i,1])).lat.values
        u_points_lats2 = np.append(u_points_lats2,u_points_lat2)
    # coordinates of v_points:
    j=0
    for j in range(len(uniques_new)):
        try:
            v_points_lon2 = v_data.isel(x=int(uniques_new[j,0]),y=int(uniques_new[j,1])+1).lon.values
        except IndexError:
            v_points_lon2 = v_data.isel(x=int(uniques_new[j,0]),y=int(uniques_new[j,1])).lon.values
        v_points_lons2 = np.append(v_points_lons2,v_points_lon2)
        try:
            v_points_lat2 = v_data.isel(x=int(uniques_new[j,0]),y=int(uniques_new[j,1])+1).lat.values
        except IndexError:
            v_points_lat2 = v_data.isel(x=int(uniques_new[j,0]),y=int(uniques_new[j,1])).lat.values
        v_points_lats2 = np.append(v_points_lats2,v_points_lat2)
    # coordinates of T_points:


    # provide line
    u_line = xa.DataArray(np.full((len(uniques_new), len(uniques_new)),1),dims=['lat','lon'],coords=[u_points_lats,u_points_lons])
    v_line = xa.DataArray(np.full((len(uniques_new), len(uniques_new)),1),dims=['lat','lon'],coords=[v_points_lats,v_points_lons])
    T_line = xa.DataArray(np.full((len(uniques_new), len(uniques_new)),1),dims=['lat','lon'],coords=[T_points_lats,T_points_lons])
    T_line2 = xa.DataArray(np.full((len(sels), len(sels)),1),dims=['lat','lon'],coords=[T_points_lats2,T_points_lons2])
    u_line2 = xa.DataArray(np.full((len(uniques_new), len(uniques_new)),1),dims=['lat','lon'],coords=[u_points_lats2,u_points_lons2])
    v_line2 = xa.DataArray(np.full((len(uniques_new), len(uniques_new)),1),dims=['lat','lon'],coords=[v_points_lats2,v_points_lons2])

    return uniques_new,u_line,v_line,T_line,u_line2,v_line2

def get_nearest_r(ref_line,u_line,v_line,T_line):
    '''
    This function calculates the nearest point on the reference line and it's neighbour for each T-point on the native grid
    '''

    mini = np.zeros((len(T_line),2)) # two columns, first for nearest point and second for neighbour
    i=0
    
    for i in tqdm(range(len(T_line))):
    #calc distance for every point of line to every point of ref line
        dist_list_single_point = []
        j=0
        for j in range(len(ref_line)):
            dist = prepro.distance(ref_line.lat[j].values,ref_line.lon[j].values,T_line.lat[i].values,T_line.lon[i].values)
            dist_list_single_point = np.append(dist_list_single_point,dist)

        minimum = dist_list_single_point.tolist().index(min(dist_list_single_point))
        #print(len(ref_line))
        #sys.exit()
        mini[i,0] = minimum     # nearest point
        if mini[i,0] == len(ref_line)-1:
            mini[i,1] = minimum-1  # left neighbouring point
        else:
            mini[i,1] = minimum+1  # right neighbouring point

    r_1_lat=[]
    r_1_lon=[]
    r_2_lat=[]
    r_2_lon=[]
    ##### select points on T-line
    for i in range(len(T_line)):
        r1lat=ref_line[int(mini[i,0]),int(mini[i,0])].lat
        r1lon=ref_line[int(mini[i,0]),int(mini[i,0])].lon
        r2lat=ref_line[int(mini[i,1]),int(mini[i,1])].lat
        r2lon=ref_line[int(mini[i,1]),int(mini[i,1])].lon
        r_1_lat=np.append(r_1_lat,r1lat)
        r_1_lon=np.append(r_1_lon,r1lon)
        r_2_lat=np.append(r_2_lat,r2lat)
        r_2_lon=np.append(r_2_lon,r2lon)
        
    return r_1_lat,r_1_lon,r_2_lat,r_2_lon

def normalize(v):
    norm = np.linalg.norm(v)
    if norm == 0: 
        return v
    return v / norm


def calc_normvec(ref_line,u_line,v_line,T_line):
    '''
    This function calculates the vectors normal to the reference line at each u,v and T point using 3 cross products
    '''
    a = 6378388
    normu_x, normu_y, normu_z = [], [], []
    normv_x, normv_y, normv_z = [], [], []
    normT_x, normT_y, normT_z = [], [], []
    
    r_1_lat,r_1_lon,r_2_lat,r_2_lon=get_nearest_r(ref_line,u_line,v_line,T_line)
    ref_x,ref_y,ref_z=prepro.kugel_2_kart(ref_line.lat.values,ref_line.lon.values)
    u_x,u_y,u_z=prepro.kugel_2_kart(u_line.lat.values,u_line.lon.values)
    v_x,v_y,v_z=prepro.kugel_2_kart(v_line.lat.values,v_line.lon.values)
    T_x,T_y,T_z=prepro.kugel_2_kart(T_line.lat.values,T_line.lon.values)
    r1_x,r1_y,r1_z=prepro.kugel_2_kart(r_1_lat,r_1_lon)
    r2_x,r2_y,r2_z=prepro.kugel_2_kart(r_2_lat,r_2_lon)
    
    Tprojx, Tprojy, Tprojz= [], [], []
    uprojx, uprojy, uprojz= [], [], []
    vprojx, vprojy, vprojz= [], [], []
    
    for i in range(len(T_x)):
        r_1=[r1_x[i],r1_y[i],r1_z[i]]
        r_2=[r2_x[i],r2_y[i],r2_z[i]]
        n=np.cross(np.array(r_1),np.array(r_2))  ## cross product 1
        #calc normvec T
        T=[T_x[i],T_y[i],T_z[i]]
        pT=np.cross(np.array(T),n)  ## cross product 2
        T_proj=np.cross(n,pT)  ## cross product 3
        T_proj=normalize(T_proj)*a  ## normalize and multiply by Earth radius to obtain point on ref line
        Tprojx=np.append(T_proj[0],Tprojx)
        Tprojy=np.append(T_proj[1],Tprojy)
        Tprojz=np.append(T_proj[2],Tprojz)
        T_t_x=T_proj[0]-T[0]   ### difference between T point and projected point on ref-line = normal vector
        T_t_y=T_proj[1]-T[1]
        T_t_z=T_proj[2]-T[2]
        normT_x=np.append(normT_x,T_t_x)
        normT_y=np.append(normT_y,T_t_y)
        normT_z=np.append(normT_z,T_t_z)
        #calc normvec u
        u=[u_x[i],u_y[i],u_z[i]]
        pu=np.cross(np.array(u),n)
        u_proj=np.cross(n,pu)
        u_proj=normalize(u_proj)*a
        uprojx=np.append(u_proj[0],uprojx)
        uprojy=np.append(u_proj[1],uprojy)
        uprojz=np.append(u_proj[2],uprojz)
        u_t_x=u_proj[0]-u[0]
        u_t_y=u_proj[1]-u[1]
        u_t_z=u_proj[2]-u[2]
        normu_x=np.append(normu_x,u_t_x)
        normu_y=np.append(normu_y,u_t_y)
        normu_z=np.append(normu_z,u_t_z)
        #calc normvec v
        v=[v_x[i],v_y[i],v_z[i]]
        pv=np.cross(np.array(v),n)
        v_proj=np.cross(n,pv)
        v_proj=normalize(v_proj)*a
        vprojx=np.append(v_proj[0],vprojx)
        vprojy=np.append(v_proj[1],vprojy)
        vprojz=np.append(v_proj[2],vprojz)
        v_t_x=v_proj[0]-v[0]
        v_t_y=v_proj[1]-v[1]
        v_t_z=v_proj[2]-v[2]
        normv_x=np.append(normv_x,v_t_x)
        normv_y=np.append(normv_y,v_t_y)
        normv_z=np.append(normv_z,v_t_z)
        
    Tproj_lat,Tproj_lon=prepro.kart_2_kugel(Tprojx,Tprojy,Tprojz)
    T_proj = xa.DataArray(np.full((len(Tproj_lat), len(Tproj_lon)),1),dims=['lat','lon'],coords=[np.flip(Tproj_lat),np.flip(Tproj_lon)])
    T_proj = T_proj.to_dataset(name='T')
    uproj_lat,uproj_lon=prepro.kart_2_kugel(uprojx,uprojy,uprojz)
    u_proj = xa.DataArray(np.full((len(uproj_lat), len(uproj_lon)),1),dims=['lat','lon'],coords=[np.flip(uproj_lat),np.flip(uproj_lon)])
    u_proj = u_proj.to_dataset(name='u')
    vproj_lat,vproj_lon=prepro.kart_2_kugel(vprojx,vprojy,vprojz)
    v_proj = xa.DataArray(np.full((len(vproj_lat), len(vproj_lon)),1),dims=['lat','lon'],coords=[np.flip(vproj_lat),np.flip(vproj_lon)])
    v_proj = v_proj.to_dataset(name='v')
    return normT_x,normT_y,normT_z, T_proj, u_proj, v_proj

def calc_dir_vector(u_line,v_line,u_line2,v_line2):
    '''
    This function calculates the direct u and v vectors
    '''

    ux,uy,uz = prepro.kugel_2_kart(u_line.lat.values,u_line.lon.values) #koordinaten uPunkt -Koordinaten TPunkt
    vx,vy,vz = prepro.kugel_2_kart(v_line.lat.values,v_line.lon.values) #koordinaten vPunkt -Koordinaten T PUnkt
    ux2,uy2,uz2 = prepro.kugel_2_kart(u_line2.lat.values,u_line2.lon.values) #koordinaten uPunkt -Koordinaten TPunkt
    vx2,vy2,vz2 = prepro.kugel_2_kart(v_line2.lat.values,v_line2.lon.values) #koordinaten vPunkt -Koordinaten T PUnkt

    rv_u = np.zeros((len(u_line),3))
    rv_v = np.zeros((len(v_line),3))
    i=0
    for i in tqdm(range(len(u_line))):
        vec_ux = ux2[i]-ux[i]#ux[i] - tx[i]#
        vec_uy = uy2[i]-uy[i]#uy[i] - ty[i]#
        vec_uz = uz2[i]-uz[i]#uz[i] - tz[i]#
        rv_u[i,0] = vec_ux
        rv_u[i,1] = vec_uy
        rv_u[i,2] = vec_uz
        
        vec_vx = vx2[i]-vx[i]#vx[j] - tx[j]#
        vec_vy = vy2[i]-vy[i]#vy[j] - ty[j]#
        vec_vz = vz2[i]-vz[i]#vz[j] - tz[j]#
        rv_v[i,0] = vec_vx
        rv_v[i,1] = vec_vy
        rv_v[i,2] = vec_vz
        
    return rv_u, rv_v

def proj_vec(ref_line,u_line,v_line,T_line,u_line2,v_line2):
    '''
    This function calculates the projection vectors usind the normal and direct vectors
    '''
    
    print('.. calculating normal vectors')
    normT_x,normT_y,normT_z,T_proj, u_proj, v_proj = calc_normvec(ref_line,u_line,v_line,T_line) #normu_x,normu_y,normu_z,normv_x,normv_y,normv_z
    print('.. calculating direct vectors')
    u_dirvec, v_dirvec = calc_dir_vector(u_line,v_line,u_line2,v_line2)
    u_dir = np.zeros((len(u_dirvec),3))
    v_dir = np.zeros((len(v_dirvec),3))
    k=0
    for k in range(len(u_dirvec)):
        u_dir[k,0] = (1/np.sqrt((u_dirvec[k,0]**2 + u_dirvec[k,1]**2 + u_dirvec[k,2]**2))) * u_dirvec[k,0]
        u_dir[k,1] = (1/np.sqrt((u_dirvec[k,0]**2 + u_dirvec[k,1]**2 + u_dirvec[k,2]**2))) * u_dirvec[k,1]
        u_dir[k,2] = (1/np.sqrt((u_dirvec[k,0]**2 + u_dirvec[k,1]**2 + u_dirvec[k,2]**2))) * u_dirvec[k,2]
        v_dir[k,0] = (1/np.sqrt((v_dirvec[k,0]**2 + v_dirvec[k,1]**2 + v_dirvec[k,2]**2))) * v_dirvec[k,0]
        v_dir[k,1] = (1/np.sqrt((v_dirvec[k,0]**2 + v_dirvec[k,1]**2 + v_dirvec[k,2]**2))) * v_dirvec[k,1]
        v_dir[k,2] = (1/np.sqrt((v_dirvec[k,0]**2 + v_dirvec[k,1]**2 + v_dirvec[k,2]**2))) * v_dirvec[k,2]

    norm_betragT = np.sqrt((normT_x**2 + normT_y**2 +normT_z**2))
    zaehler_u = (u_dir[:,0] * normT_x) + (u_dir[:,1] * normT_y) + (u_dir[:,2] * normT_z)
    zaehler_v = (v_dir[:,0] * normT_x) + (v_dir[:,1] * normT_y) + (v_dir[:,2] * normT_z)
    term1_u = zaehler_u/norm_betragT**2 
    term1_v = zaehler_v/norm_betragT**2
    proj_u = np.zeros((len(normT_x),3))
    proj_v = np.zeros((len(normT_x),3))
    for i in range(len(normT_x)):
        proj_u[i,0] = term1_u[i] *normT_x[i]
        proj_u[i,1] = term1_u[i] *normT_y[i]
        proj_u[i,2] = term1_u[i] *normT_z[i]
        proj_v[i,0] = term1_v[i] *normT_x[i]
        proj_v[i,1] = term1_v[i] *normT_y[i]
        proj_v[i,2] = term1_v[i] *normT_z[i]

    return proj_u,proj_v,T_proj, u_proj, v_proj

def calc_dx(T_proj_points):
    
    dist_listT = []
    for k in range(len(T_line)-1):
        distsT = prepro.distance(T_proj_points.lat[k],T_proj_points.lon[k],T_proj_points.lat[k+1],T_proj_points.lon[k+1])
        dist_listT = np.append(dist_listT,distsT)
    dist_lastT = prepro.distance(T_proj_points.lat[-2],T_proj_points.lon[-2],T_proj_points.lat[-1],T_proj_points.lon[-1])
    dist_listT = np.append(dist_listT,dist_lastT)

    T_proj_lat=T_proj_points.lat*dist_listT/dist_listT
    T_proj_lat=T_proj_lat[~np.isnan(T_proj_lat)]
    T_proj_lon=T_proj_points.lon*dist_listT/dist_listT
    T_proj_lon=T_proj_lon[~np.isnan(T_proj_lon)]
    
    dist_listT_kurz2 = []
    dist_firstT_kurz2 = prepro.distance(T_proj_lat[0],T_proj_lon[0],T_proj_lat[1],T_proj_lon[1])
    dist_listT_kurz2 = np.append(dist_listT_kurz2,dist_firstT_kurz2)
    for k in range(1,len(T_proj_lon)-1):      
        distsT_kurz2 = prepro.distance(T_proj_lat[k-1],T_proj_lon[k-1],T_proj_lat[k],T_proj_lon[k])/2 + prepro.distance(T_proj_lat[k],T_proj_lon[k],T_proj_lat[k+1],T_proj_lon[k+1])/2
        dist_listT_kurz2 = np.append(dist_listT_kurz2,distsT_kurz2)
    dist_lastT_kurz2 = prepro.distance(T_proj_lat[-2],T_proj_lon[-2],T_proj_lat[-1],T_proj_lon[-1])
    dist_listT_kurz2 = np.append(dist_listT_kurz2,dist_lastT_kurz2)

    return dist_listT,dist_listT_kurz2,T_proj_lat, T_proj_lon

def calc_betrag(projvar):
    betrag = np.sqrt((projvar[:,0]**2 + projvar[:,1]**2 + projvar[:,2]**2))
    return betrag

########### Hier ev. noch verbessern!!!!

def multi_factors(T_line,u_line,v_line,T_proj,u_proj,v_proj):
    T_dist=prepro.distance(T_line.lat.values+1,T_line.lon.values+1,T_proj.lat.values,T_proj.lon.values)
    u_dist=prepro.distance(u_line.lat.values+1,u_line.lon.values+1,u_proj.lat.values,u_proj.lon.values)
    v_dist=prepro.distance(v_line.lat.values+1,v_line.lon.values+1,v_proj.lat.values,v_proj.lon.values)
    tu=u_dist-T_dist
    tv=v_dist-T_dist
    tu[tu<0]=-1
    tu[tu>0]=1
    tv[tv<0]=-1
    tv[tv>0]=1
    
    return tu,tv

def calc_interpolation_points(indices,T_data, ref_line):
    '''
    This function selects indices on T,u and v grid
    args:
        path_indices: path with indices files
        strait: Fram,Barents,GSR,etc.
        path_raw_data: path to raw fata files
        reana: ORAS5,FOAM,CGLORS,GLORYS2V4
    '''
    selection = indices.indices.values
    sels = np.zeros((len(selection),2))
    i=0
    for i in range(len(selection)):
        if selection[i,0] == 0:
            sels[i,0] = selection[i,2]
            sels[i,1] = selection[i,3]
        else:
            sels[i,0] = selection[i,0]
            sels[i,1] = selection[i,1]


    # prepate for coordinates
    T_points_lons = []
    T_points_lats = []
    # coordinates of T_points:
    k=0
    for k in range(len(sels)):
        T_points_lon = T_data.isel(x=int(sels[k,0]),y=int(sels[k,1])).lon.values
        T_points_lons = np.append(T_points_lons,T_points_lon)
        T_points_lat = T_data.isel(x=int(sels[k,0]),y=int(sels[k,1])).lat.values
        T_points_lats = np.append(T_points_lats,T_points_lat)
    T_line = xa.DataArray(np.full((len(sels), len(sels)),1),dims=['lat','lon'],coords=[T_points_lats,T_points_lons])

    mini = np.zeros((len(T_line),2)) # two columns, first for nearest point and second for neighbour
    i=0
    for i in range(len(T_line)):
    #calc distance for every point of line to every point of ref line
        dist_list_single_point = []
        j=0
        for j in range(len(ref_line)):
            dist = prepro.distance(ref_line.lat[j].values,ref_line.lon[j].values,T_line.lat[i].values,T_line.lon[i].values)
            dist_list_single_point = np.append(dist_list_single_point,dist)
        minimum = dist_list_single_point.tolist().index(min(dist_list_single_point))

        mini[i,0] = minimum     # nearest point
        if mini[i,0] == len(ref_line)-1:
            mini[i,1] = minimum-1  # left neighbouring point
        else:
            mini[i,1] = minimum+1  # right neighbouring point


    r_1_lat=[]
    r_1_lon=[]
    r_2_lat=[]
    r_2_lon=[]
    ##### select points on T-line
    for i in range(len(T_line)):
        r1lat=ref_line[int(mini[i,0]),int(mini[i,0])].lat
        r1lon=ref_line[int(mini[i,0]),int(mini[i,0])].lon
        r2lat=ref_line[int(mini[i,1]),int(mini[i,1])].lat
        r2lon=ref_line[int(mini[i,1]),int(mini[i,1])].lon
        r_1_lat=np.append(r_1_lat,r1lat)
        r_1_lon=np.append(r_1_lon,r1lon)
        r_2_lat=np.append(r_2_lat,r2lat)
        r_2_lon=np.append(r_2_lon,r2lon)

    a = 6378388
    normT_x=[]
    normT_y=[]
    normT_z=[]
    ref_x,ref_y,ref_z=prepro.kugel_2_kart(ref_line.lat.values,ref_line.lon.values)
    T_x,T_y,T_z=prepro.kugel_2_kart(T_line.lat.values,T_line.lon.values)
    r1_x,r1_y,r1_z=prepro.kugel_2_kart(r_1_lat,r_1_lon)
    r2_x,r2_y,r2_z=prepro.kugel_2_kart(r_2_lat,r_2_lon)
    Tprojx=[]
    Tprojy=[]
    Tprojz=[]
    for i in range(len(T_x)):
        r_1=[r1_x[i],r1_y[i],r1_z[i]]
        r_2=[r2_x[i],r2_y[i],r2_z[i]]
        n=np.cross(np.array(r_1),np.array(r_2))  ## cross product 1
        #calc normvec T
        T=[T_x[i],T_y[i],T_z[i]]
        pT=np.cross(np.array(T),n)  ## cross product 2
        T_proj=np.cross(n,pT)  ## cross product 3
        T_proj=normalize(T_proj)*a  ## normalize and multiply by Earth radius to obtain point on ref line
        Tprojx=np.append(T_proj[0],Tprojx)
        Tprojy=np.append(T_proj[1],Tprojy)
        Tprojz=np.append(T_proj[2],Tprojz)

    res,ind = np.unique(Tprojx, return_index=True)
    Tprojx = res[np.argsort(ind)]
    res,ind = np.unique(Tprojy, return_index=True)
    Tprojy = res[np.argsort(ind)]
    res,ind = np.unique(Tprojz, return_index=True)
    Tprojz = res[np.argsort(ind)]
    
    
    Tproj_lat,Tproj_lon=prepro.kart_2_kugel(Tprojx,Tprojy,Tprojz)
    T_proj = xa.DataArray(np.full((len(Tproj_lat), len(Tproj_lon)),1),dims=['lat','lon'],coords=[np.flip(Tproj_lat),np.flip(Tproj_lon)])
    T_proj = T_proj.to_dataset(name='T')

    dist_listT = []
    for k in range(len(Tprojx)-1):
        distsT = prepro.distance(T_proj.lat[k],T_proj.lon[k],T_proj.lat[k+1],T_proj.lon[k+1])
        dist_listT = np.append(dist_listT,distsT)
    dist_lastT = prepro.distance(T_proj.lat[-2],T_proj.lon[-2],T_proj.lat[-1],T_proj.lon[-1])
    dist_listT = np.append(dist_listT,dist_lastT)

    T_proj_lat=T_proj.lat*dist_listT/dist_listT
    T_proj_lat=T_proj_lat[~np.isnan(T_proj_lat)]
    T_proj_lon=T_proj.lon*dist_listT/dist_listT
    T_proj_lon=T_proj_lon[~np.isnan(T_proj_lon)]
    
    dist_listT_kurz2 = []
    dist_firstT_kurz2 = prepro.distance(T_proj_lat[0],T_proj_lon[0],T_proj_lat[1],T_proj_lon[1])
    dist_listT_kurz2 = np.append(dist_listT_kurz2,dist_firstT_kurz2)
    for k in range(1,len(T_proj_lon)-1):      
        distsT_kurz2 = prepro.distance(T_proj_lat[k-1],T_proj_lon[k-1],T_proj_lat[k],T_proj_lon[k])/2 + prepro.distance(T_proj_lat[k],T_proj_lon[k],T_proj_lat[k+1],T_proj_lon[k+1])/2
        dist_listT_kurz2 = np.append(dist_listT_kurz2,distsT_kurz2)
    dist_lastT_kurz2 = prepro.distance(T_proj_lat[-2],T_proj_lon[-2],T_proj_lat[-1],T_proj_lon[-1])
    dist_listT_kurz2 = np.append(dist_listT_kurz2,dist_lastT_kurz2)

    return T_proj,dist_listT*1000,dist_listT_kurz2*1000







import os
import xarray as xa
import numpy as np
import sys

def check_res(Tdataset):
    if np.abs(Tdataset.lon.min()-Tdataset.lon.max()) >= 358:
        res=360/len(Tdataset.x)*0.4
    else:
        res=(Tdataset.lon[0,1].values-Tdataset.lon[0,0].values)*0.4
    print(res)
    return res


def def_indices(strait,coords,lon_p,lat_p,set_latlon,res):
    if set_latlon == True:
        lon = lon_p
        lat = lat_p
    elif coords!=0:
        if np.abs(coords[2]-coords[0]) <= np.abs(coords[3]-coords[1]):
            lon = np.array(np.arange(coords[1],coords[3],res))
            lat = np.array(np.linspace(coords[0],coords[2],len(lon)))
        else:
            lat = np.array(np.arange(coords[0],coords[2],res))
            lon = np.array(np.linspace(coords[1],coords[3],len(lat)))
    elif strait == 'Bering':
        lon = np.array(np.arange(-170.5,-166,res))
        lat = np.array(np.linspace(65.99,65.75,len(lon)))
    elif strait == 'Fram':
        lon = np.array(np.arange(-20.7,12,res))
        lat = np.array(np.linspace(78.82,78.836,len(lon)))
    elif strait == 'Davis':
        lon = np.array(np.arange(-61.8,-52.5,res))
        lat = np.array(np.linspace(66.65,67.31,len(lon)))
    elif strait == 'Barents':
        lat = np.array(np.arange(78,69.2,-res))
        lon = np.array(np.linspace(18,19.8,len(lat)))
    elif strait == 'RAPID':
        lon = np.array(np.arange(-80.5,-13.5,res))
        lat = np.array(np.linspace(26.0,26.0,len(lon)))
    elif strait == 'OSNAP':
        lon1 = np.array(np.arange(-57,-44.6,res))
        lat1 = np.array(np.linspace(52,60.1888,len(lon1)))
        lon2 = np.array(np.arange(-44.6,-5.5,res))
        lat2 = np.array(np.linspace(60.1888,56.5,len(lon2)))
        lat = np.append(lat1,lat2)
        lon = np.append(lon1,lon2)
    elif strait == 'GSR':
        lon1 = np.array(np.arange(-30.8212,-23.2447,res))
        lon2 = np.array(np.arange(-23.2447,-15.0679,res))
        lon3 = np.array(np.arange(-15.0679,-6.8700,res))
        lon4 = np.array(np.arange(-6.8700,-1.1695,res))
        lon5 = np.array(np.arange(-1.1695,6.10547,res))
        lon = []
        lon = np.append(lon,lon1)
        lon = np.append(lon,lon2)
        lon = np.append(lon,lon3)
        lon = np.append(lon,lon4)
        lon = np.append(lon,lon5)
        lat1 = np.array(np.linspace(68.5318,66.0050,len(lon1)))
        lat2 = np.array(np.linspace(66.0050,64.4121,len(lon2)))
        lat3 = np.array(np.linspace(64.4121,62.0669,len(lon3)))
        lat4 = np.array(np.linspace(62.0669,60.2777,len(lon4)))
        lat5 = np.array(np.linspace(60.2777,59.4703,len(lon5)))
        lat = []
        lat = np.append(lat,lat1)
        lat = np.append(lat,lat2)
        lat = np.append(lat,lat3)
        lat = np.append(lat,lat4)
        lat = np.append(lat,lat5)
    elif strait == 'Makassar':
        lon = np.array(np.arange(116,119,res))
        lat = np.array(np.linspace(-2.87,-2.87,len(lon)))
    elif strait == 'Hudson':
        lon = np.array(np.arange(-82.832771,-82.59107,res))
        lat = np.array(np.linspace(69.66432,69.89028,len(lon)))
    elif strait == 'Hudson_NW':
        lon = np.array(np.arange(-84.500731,-84.317820,res))
        lat = np.array(np.linspace(69.85032,70.001132,len(lon)))
    elif strait == 'Färöer':
        lat = np.array(np.arange(62.5,63.875,res))
        lon = np.array(np.linspace(-6.125,-6.125,len(lat)))
    elif strait == 'Gibraltar':
        lat = np.array(np.arange(34.5,37.5,res))
        lon = np.array(np.linspace(-5.59,-5.59,len(lat)))
    elif strait == 'NIIC':
        lon = np.array(np.arange(-30.8212,-23.2447,res))
        lat = np.array(np.linspace(68.5318,66.0050,len(lon)))
    elif strait == 'IF':
        lon = np.array(np.arange(-15.0679,-6.8700,res))
        lat = np.array(np.linspace(64.4121,62.0669,len(lon)))
    elif strait == 'FS':
        lon = np.array(np.arange(-6.8700,-1.1695,res))
        lat = np.array(np.linspace(62.0669,60.2777,len(lon)))
    else:
        print('strait not defined, please provide coordinates: coords=(lat_start,lon_start,lat_end,lon_end)')
        sys.exit()

    return lat,lon


def distance(lat1,lon1,lat2,lon2):
    '''
    This funtion provides the distance calculation on the sphere surface (source: https://www.kompf.de/gps/distcalc.html)
    args:
        lat1,lon1 = pair of values on reference line
        lat2,lat2 = pair of values for which the distance to the reference line should be tested
    returns: 
        distance of (lat1,lon1) and (lat2,lon2) in kilometers
    '''
    lat11 = np.radians(lat1)
    lon11 = np.radians(lon1)
    lat22 = np.radians(lat2)
    lon22 = np.radians(lon2)
    a = 6378.388
    part1 = np.sin(lat11)*np.sin(lat22) 
    part2 = np.cos(lat11)*np.cos(lat22)*np.cos(lon22-lon11)
    term = part1+part2
    # following if/else statement is necessary due to numerical issues
    if type(term) != float:   
        term = np.where(term>1,1,term)
    else:
        if term >1:
            term=1
    distance = a * np.arccos(term)
    return distance

def selection_window(lat1,lon1,lat2,lon2,Tdataset):
    '''
    This funtion selects the nearest point on Tgrid within a selection window with a size of +- 0.5 degrees around the first point on the reference curve.
    args:
         Tdataset: xa.Dataset of Tdata on Tgrid 
         lat1,lon1 = pair of values on reference line
         lat2,lat2 = pair of values for which the distance to the reference line should be tested
    returns:
         xloc: index of nearest x on Tgrid
         yloc: index of nearest y on Tgrid
    '''
    delta=2
    x1_x2 = np.where((lat2>(lat1-delta))&(lat2<(lat1+delta)),1,0)
    y1_y2 = np.where((lon2>(lon1-delta))&(lon2<(lon1+delta)),1,0)
    fenster = x1_x2 + y1_y2
    fenster = np.where(fenster==2,1,0)
    lat2 = (lat2 * fenster)
    lon2 = (lon2 * fenster)
    dist1 = distance(lat1,lon1,lat2,lon2)

    if Tdataset.lat.dims[0] == 'x':
        xloc,yloc = np.where(dist1==np.nanmin(dist1))
    else:
        yloc,xloc = np.where(dist1==np.nanmin(dist1))

    yloc=yloc[0]
    xloc=xloc[0]
    return xloc, yloc

def select_points(Tdataset,xstart,ystart,lat_line_point,lon_line_point):
    '''
    This function test the neighboring grid boxes which index is the next one.
    args:
        Tdataset: xa.Dataset of Tdata on Tgrid
        xstart: x index of point from where the neighboring boxes have to be checked
        ystart: y index of point from where the neighhoring boxes have to be checked
        lat_line_point: latitude of point at reference curve
        lon_line_point: longitude of point at reference curve
    returns:
        xstart: x index of new next point
        ystart: y index of new next point
    '''
    # select adjoining points/gridboxes
    mid = Tdataset.isel(x=xstart,y=ystart)
    if ystart >= len(Tdataset.y)-1:
        above = Tdataset.isel(x=xstart,y=ystart)
    else:
        above = Tdataset.isel(x=xstart,y=ystart+1)
    under = Tdataset.isel(x=xstart,y=ystart-1)
    left = Tdataset.isel(x=xstart-1,y=ystart)
    right = Tdataset.isel(x=xstart+1,y=ystart)
    # calculate distance:
    mid_dist = distance(lat_line_point,lon_line_point,mid.lat.values,mid.lon.values)
    abo_dist =  distance(lat_line_point,lon_line_point,above.lat.values,above.lon.values)
    und_dist =  distance(lat_line_point,lon_line_point,under.lat.values,under.lon.values)
    lef_dist =  distance(lat_line_point,lon_line_point,left.lat.values,left.lon.values)
    rig_dist =  distance(lat_line_point,lon_line_point,right.lat.values,right.lon.values)
    # select point with minmal distance:
    testlist = list([mid_dist,abo_dist,und_dist,lef_dist,rig_dist])

    min_pos = testlist.index(min(testlist))
    
    if min_pos == 0:
        xstart,ystart = xstart,ystart
    elif min_pos ==1:
        xstart,ystart = xstart,ystart+1
    elif min_pos ==2:
        xstart,ystart = xstart,ystart-1
    elif min_pos ==3:
        xstart,ystart = xstart-1,ystart
    elif min_pos ==4:
        xstart,ystart = xstart+1,ystart
    return xstart,ystart


def check_availability_indices(Tdataset,strait,model,coords,lon_p,lat_p,set_latlon):
    print('calculating indices...')
    res = check_res(Tdataset)
    lat,lon = def_indices(strait,coords,lon_p,lat_p,set_latlon,res)


    line = xa.DataArray(data=1,dims=['lat','lon'],coords=[lat,lon])


    # first point of line:
    start_lat, start_lon = line.isel(lat=0,lon=0).lat.values, line.isel(lat=0,lon=0).lon.values
    indices = np.zeros((len(line.lon.values),3))######
    for i in range(len(line.lon.values)):#####
        if i == 0:
            xstart,ystart =selection_window(start_lat,start_lon,Tdataset.lat.values,Tdataset.lon.values,Tdataset)
            indices[i,0] = xstart
            indices[i,1] = ystart
        else:
            xstart,ystart = select_points(Tdataset,xstart,ystart,line.lat[i].values,line.lon[i].values)
            # to provide circle: 
            if xstart == len(Tdataset.x)-1:
                xstart = 0
            elif ystart == len(Tdataset.y)-1 and indices[i-1,1] == len(Tdataset.y)-1:
                print('Attention: Strait crossing the northern boundary – make sure correct indices are chosen!')
                ystart = len(Tdataset.y)-1
                xstart = len(Tdataset.x)-1-xstart
            indices[i,0] = xstart
            indices[i,1] = ystart

    # remove duplicates:
    row_mask = np.append([True],np.any(np.diff(indices,axis=0),1))
    out = indices[row_mask]

    # if first and last point of out are the same:
    if out[0,0] == out[-1,0] and out[0,1] == out[-1,1]:
        out = np.delete(out,-1,0)

    # allocting if a u point or a v point are taken:
    selection = np.zeros((len(out),5))
    for i in range(1,len(out)): # starting at 1 because point at i-1 is checked to see how point at i is reached
        # came from left:
        if out[i-1,0]+1 == out[i,0] and out[i-1,1] == out[i,1]:
            if i == 1:
                selection[i-1,2] = out[i-1,0]
                selection[i-1,3] = out[i-1,1]
                selection[i-i,4] = 1
            selection[i,2] = out[i,0]
            selection[i,3] = out[i,1]
            selection[i,4] = 1

        # came from right:
        elif out[i-1,0]-1 == out[i,0] and out[i-1,1] == out[i,1]:
            if i ==1:
                selection[i-1,2] = out[i-1,0] +1
                selection[i-1,3] = out[i-1,1]
                selection[i-1,4] =1
            selection[i,2] = out[i,0] +1
            selection[i,3] = out[i,1]
            selection[i,4] = 1

        # came from above:
        elif out[i-1,0] == out[i,0] and out[i-1,1]-1 == out[i,1]:
            if i == 1:
                selection[i-1,0] = out[i-1,0]
                selection[i-1,1] = out[i-1,1]
                selection[i-1,4] =1
            selection[i,0] = out[i-1,0]
            selection[i,1] = out[i-1,1]
            selection[i,4] =1
    
        # came from below:
        elif out[i-1,0] == out[i,0] and out[i-1,1]+1 == out[i,1]:
            if i == 1:
                selection[i-1,0] = out[i-1,0]
                selection[i-1,1] = out[i-1,1]
                selection[i-1,4] = -1
            selection[i,0] = out[i,0]
            selection[i,1] = out[i,1]
            selection[i,4] = -1

        # y is getting smaller:
        elif out[i-1,1] == out[i,1]+1:
            selection[i,0] = out[i,0]
            selection[i,1] = out[i,1]
            selection[i,4] = -1
    
        # x index is changing:
        elif out[i-1,0] > 1000 and out[i,0] < 1000:
            selection[i,2] = out[i,0]
            selection[i,3] = out[i,1]
            selection[i,4] = 1
        else:
            ('hmmm...')


       
    row_mask = np.append([True],np.any(np.diff(selection,axis=0),1))
    selection = selection[row_mask]
    res,ind = np.unique(selection, return_index=True, axis=0)
    selection = res[np.argsort(ind)]

    # saving choosen indices
    selected = xa.Dataset({'indices':(('dim0','dim1'),selection)})
    selected.indices.attrs['dim0'] = 'points'
    selected.indices.attrs['dim1'] = 'columns with points on ugrid,vgrid and sign'
    selected.indices.attrs['dim1_general'] = 'if values are 0 then on the other grid'
    selected.indices.attrs['dim1=0'] = 'selected x index on ugrid'
    selected.indices.attrs['dim1=1'] = 'selected y index on ugrid'
    selected.indices.attrs['dim1=2'] = 'selected x index on vgrid'
    selected.indices.attrs['dim1=3'] = 'selected y index on vgrid'
    selected.indices.attrs['dim1=4'] = 'sign concerning ugrid (when it has to be counted negative)'

    return selected,line



def unique_rows(a):
    a = np.ascontiguousarray(a)
    unique_a = np.unique(a.view([('',a.dtype)]*a.shape[1]))
    return unique_a.view(a.dtype).reshape(unique_a.shape[0],a.shape[1])


def prepare_indices(indices):
    '''
    This function prepares the points when ugrid or vgrid have to be choosen.
    args:
        indices: xa.Dataset where the indices where saved
    returns:
        out_u: np.array with indices where ugrid needs to be selected
        out_v: np.array with indices where vgrid needs to be selected
        out_u_vz: np.array with signs for the indices on ugrid
    '''
    select = indices.indices.values
    selection = unique_rows(select)
    out_u = selection[:,0:2]
    masku = np.all(np.isnan(out_u)|np.equal(out_u,0),axis=1)
    out_u = out_u[~masku]
    out_v = selection[:,2:4]
    maskv = np.all(np.isnan(out_v)|np.equal(out_v,0),axis=1)
    out_v = out_v[~maskv]
    out_u_vz = np.delete(selection,2,1)
    out_u_vz = np.delete(out_u_vz,2,1)
    out_u_vz = out_u_vz[~masku]

    return out_u,out_v,out_u_vz

